from gevent import monkey

monkey.patch_all()

import logging
import os
import platform
import sqlite3
import sys
import traceback
from datetime import datetime, timezone
from logging.handlers import TimedRotatingFileHandler

import colorlog
import flask_wtf
import pika
import pytz
import requests
import sqlalchemy
import yaml
from flask import Flask, current_app, g, request, session
from flask_cors import CORS
from flask_migrate import Migrate, upgrade
from flask_security import (
    Security,
    SQLAlchemyUserDatastore,
    hash_password,
    uia_email_mapper,
    uia_username_mapper,
)
from flask_security.models import fsqla_v3
from flask_security.models import fsqla_v3 as fsqla
from flask_security.signals import user_registered
from sqlalchemy import insert
from werkzeug.middleware.proxy_fix import ProxyFix

import opentakserver
from opentakserver.certificate_authority import CertificateAuthority
from opentakserver.controllers.meshtastic_controller import MeshtasticController
from opentakserver.defaultconfig import DefaultConfig
from opentakserver.EmailValidator import EmailValidator
from opentakserver.extensions import apscheduler, babel, db, ldap_manager, logger, mail, socketio
from opentakserver.models.Group import Group, GroupTypeEnum
from opentakserver.models.Icon import Icon
from opentakserver.models.role import Role
from opentakserver.models.WebAuthn import WebAuthn
from opentakserver.PasswordValidator import PasswordValidator
from opentakserver.plugins.Plugin import Plugin
from opentakserver.plugins.PluginManager import PluginManager
from opentakserver.sql_jobstore import SQLJobStore
from opentakserver.UsernameValidator import UsernameValidator

try:
    from opentakserver.mumble.mumble_ice_app import MumbleIceDaemon
except ModuleNotFoundError:
    print("Mumble auth not supported on this platform")


def get_locale():
    if "language" in session:
        return session["language"]
    return request.accept_languages.best_match(current_app.config.get("OTS_LANGUAGES").keys())


def get_timezone():
    # Always return UTC and let the frontend handle converting timezones
    return pytz.timezone("UTC")


def init_extensions(app):
    db.init_app(app)
    Migrate(app, db)

    logger.info(f"OpenTAKServer {opentakserver.__version__}")
    logger.info("Loading the database...")
    with app.app_context():
        upgrade(
            directory=os.path.join(
                os.path.dirname(os.path.realpath(opentakserver.__file__)), "migrations"
            )
        )
        # Flask-Migrate does weird things to the logger
        logger.disabled = False
        logger.parent.handlers.pop()
        if app.config.get("DEBUG"):
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    # Handle config options that can't be serialized to yaml
    app.config.update(
        {
            "SCHEDULER_JOBSTORES": {
                "default": SQLJobStore(url=app.config.get("SQLALCHEMY_DATABASE_URI"))
            }
        }
    )
    identity_attributes = [{"username": {"mapper": uia_username_mapper, "case_insensitive": True}}]

    # Don't allow registration unless email is enabled
    if app.config.get("OTS_ENABLE_EMAIL"):
        identity_attributes.append(
            {"email": {"mapper": uia_email_mapper, "case_insensitive": True}}
        )
        app.config.update(
            {
                "SECURITY_REGISTERABLE": True,
                "SECURITY_CONFIRMABLE": True,
                "SECURITY_RECOVERABLE": True,
                "SECURITY_TWO_FACTOR_ENABLED_METHODS": ["authenticator", "email"],
            }
        )
    else:
        app.config.update(
            {
                "SECURITY_REGISTERABLE": False,
                "SECURITY_CONFIRMABLE": False,
                "SECURITY_RECOVERABLE": False,
                "SECURITY_TWO_FACTOR_ENABLED_METHODS": ["authenticator"],
            }
        )

    if app.config.get("OTS_ENABLE_LDAP"):
        logger.info("Enabling LDAP")
        ldap_manager.init_app(app)
        identity_attributes.append({"ldap": {}})

    app.config.update({"SECURITY_USER_IDENTITY_ATTRIBUTES": identity_attributes})

    ca = CertificateAuthority(logger, app)
    ca.create_ca()

    cors = CORS(
        app,
        resources={
            r"/api/*": {"origins": "*"},
            r"/Marti/*": {"origins": "*"},
            r"/*": {"origins": "*"},
        },
        supports_credentials=True,
    )
    flask_wtf.CSRFProtect(app)

    socketio_logger = False
    if app.config.get("DEBUG"):
        socketio_logger = logger
    socketio.init_app(
        app,
        logger=socketio_logger,
        ping_timeout=1,
        message_queue="amqp://" + app.config.get("OTS_RABBITMQ_SERVER_ADDRESS"),
    )

    rabbit_credentials = pika.PlainCredentials(
        app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD")
    )
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(
        pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials)
    )

    channel = rabbit_connection.channel()
    channel.exchange_declare("dms", durable=True, exchange_type="direct")
    channel.exchange_declare("cot_parser", durable=True, exchange_type="direct")
    channel.exchange_declare("chatrooms", durable=True, exchange_type="direct")
    channel.exchange_declare(
        "missions", durable=True, exchange_type="topic"
    )  # For Data Sync mission feeds
    channel.exchange_declare("groups", durable=True, exchange_type="topic")  # For channels/groups
    channel.exchange_declare(
        "firehose", durable=True, exchange_type="fanout"
    )  # A firehose of all CoT data
    channel.close()
    rabbit_connection.close()

    if not apscheduler.running:
        apscheduler.init_app(app)
        apscheduler.start(paused=False)

    try:
        fsqla.FsModels.set_db_info(db)
    except sqlalchemy.exc.InvalidRequestError:
        pass

    from opentakserver.models.role import Role
    from opentakserver.models.user import User

    user_datastore = SQLAlchemyUserDatastore(db, User, Role, WebAuthn)
    app.security = Security(
        app,
        user_datastore,
        mail_util_cls=EmailValidator,
        password_util_cls=PasswordValidator,
        username_util_cls=UsernameValidator,
    )

    mail.init_app(app)

    babel.init_app(app, locale_selector=get_locale, timezone_selector=get_timezone)


def setup_logging(app):
    level = logging.INFO
    if app.config.get("DEBUG"):
        level = logging.DEBUG
    logger.setLevel(level)

    if sys.stdout.isatty():
        color_log_handler = colorlog.StreamHandler()
        color_log_formatter = colorlog.ColoredFormatter(
            "%(log_color)s[%(asctime)s] - OpenTAKServer[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S %Z",
        )
        color_log_handler.setFormatter(color_log_formatter)
        logger.addHandler(color_log_handler)
        logger.info("Added color logger")

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs"), exist_ok=True)
    fh = TimedRotatingFileHandler(
        os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs", "opentakserver.log"),
        when=app.config.get("OTS_LOG_ROTATE_WHEN"),
        interval=app.config.get("OTS_LOG_ROTATE_INTERVAL"),
        backupCount=app.config.get("OTS_BACKUP_COUNT"),
    )
    fh.setFormatter(
        logging.Formatter(
            "[%(asctime)s] - OpenTAKServer[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s"
        )
    )
    logger.addHandler(fh)


def create_app(cli=True):
    app = Flask(__name__)
    app.config.from_object(DefaultConfig)
    setup_logging(app)

    if not cli:
        # Load config.yml if it exists
        if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
            app.config.from_file(
                os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load
            )
        else:
            # First run, created config.yml based on default settings
            logger.info("Creating config.yml")
            with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
                conf = {}
                for option in DefaultConfig.__dict__:
                    # Fix the sqlite DB path on Windows
                    if (
                        option == "SQLALCHEMY_DATABASE_URI"
                        and platform.system() == "Windows"
                        and DefaultConfig.__dict__[option].startswith("sqlite")
                    ):
                        conf[option] = (
                            DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
                        )
                    elif option.isupper():
                        conf[option] = DefaultConfig.__dict__[option]
                config.write(yaml.safe_dump(conf))

        # Try to set the MediaMTX token
        if app.config.get("OTS_MEDIAMTX_ENABLE"):
            try:
                new_conf = None
                with open(
                    os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "mediamtx.yml"), "r"
                ) as mediamtx_config:
                    conf = mediamtx_config.read()
                    if "MTX_TOKEN" in conf:
                        new_conf = conf.replace("MTX_TOKEN", app.config.get("OTS_MEDIAMTX_TOKEN"))
                if new_conf:
                    with open(
                        os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "mediamtx.yml"),
                        "w",
                    ) as mediamtx_config:
                        mediamtx_config.write(new_conf)
            except BaseException as e:
                logger.error("Failed to set MediaMTX token: {}".format(e))
        else:
            logger.info("MediaMTX disabled")

        init_extensions(app)

        from opentakserver.blueprints.marti_api import marti_blueprint

        app.register_blueprint(marti_blueprint)

        from opentakserver.blueprints.ots_api import ots_api

        app.register_blueprint(ots_api)

        from opentakserver.blueprints.ots_socketio import ots_socketio_blueprint

        app.register_blueprint(ots_socketio_blueprint)

        from opentakserver.blueprints.scheduled_jobs import scheduler_blueprint

        app.register_blueprint(scheduler_blueprint)

        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_host=1)

    else:
        from opentakserver.blueprints.cli import ots, translate

        app.cli.add_command(ots, name="ots")
        app.cli.add_command(translate, name="translate")

        if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
            app.config.from_file(
                os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load
            )
            db.init_app(app)
            Migrate(app, db)

        flask_wtf.CSRFProtect(app)

        try:
            fsqla.FsModels.set_db_info(db)
        except sqlalchemy.exc.InvalidRequestError:
            pass

        from opentakserver.models.role import Role
        from opentakserver.models.user import User

        user_datastore = SQLAlchemyUserDatastore(db, User, Role, WebAuthn)
        app.security = Security(
            app,
            user_datastore,
            mail_util_cls=EmailValidator,
            password_util_cls=PasswordValidator,
            username_util_cls=UsernameValidator,
        )

        # Register blueprints to properly import all the DB models without circular imports
        from opentakserver.blueprints.marti_api import marti_blueprint

        app.register_blueprint(marti_blueprint)

        from opentakserver.blueprints.ots_api import ots_api

        app.register_blueprint(ots_api)

        from opentakserver.blueprints.ots_socketio import ots_socketio_blueprint

        app.register_blueprint(ots_socketio_blueprint)

        from opentakserver.blueprints.scheduled_jobs import scheduler_blueprint

        app.register_blueprint(scheduler_blueprint)

    return app


def create_default_groups(app):
    with app.app_context():
        if not app.config.get("OTS_ENABLE_LDAP"):
            anon_group = db.session.execute(
                db.session.query(Group).filter_by(name="__ANON__")
            ).first()
            adsb_group = db.session.execute(
                db.session.query(Group).filter_by(name=app.config.get("OTS_ADSB_GROUP"))
            ).first()
            ais_group = db.session.execute(
                db.session.query(Group).filter_by(name=app.config.get("OTS_AIS_GROUP"))
            ).first()
            meshtastic_group = db.session.execute(
                db.session.query(Group).filter_by(name=app.config.get("OTS_MESHTASTIC_GROUP"))
            ).first()

            # Commit to DB after every one to ensure that get_next_bitpos works

            if not anon_group:
                logger.info("Creating the __ANON__ group")
                anon_group = Group()
                anon_group.name = "__ANON__"
                anon_group.type = GroupTypeEnum.SYSTEM
                anon_group.bitpos = 2
                db.session.add(anon_group)
                db.session.commit()

            if not adsb_group:
                logger.info(f"Creating the {app.config.get('OTS_ADSB_GROUP')} group")
                adsb_group = Group()
                adsb_group.name = app.config.get("OTS_ADSB_GROUP")
                adsb_group.type = GroupTypeEnum.SYSTEM
                adsb_group.bitpos = adsb_group.get_next_bitpos()
                db.session.add(adsb_group)
                db.session.commit()

            if not ais_group:
                logger.info(f"Creating the {app.config.get('OTS_AIS_GROUP')} group")
                ais_group = Group()
                ais_group.name = app.config.get("OTS_AIS_GROUP")
                ais_group.type = GroupTypeEnum.SYSTEM
                ais_group.bitpos = ais_group.get_next_bitpos()
                db.session.add(ais_group)
                db.session.commit()

            if not meshtastic_group:
                logger.info(f"Creating the {app.config.get('OTS_MESHTASTIC_GROUP')} group")
                meshtastic_group = Group()
                meshtastic_group.name = app.config.get("OTS_MESHTASTIC_GROUP")
                meshtastic_group.type = GroupTypeEnum.SYSTEM
                meshtastic_group.bitpos = meshtastic_group.get_next_bitpos()
                db.session.add(meshtastic_group)
                db.session.commit()


def main(app):
    with app.app_context():
        # Download the icon sets if they aren't already in the DB
        icons = db.session.query(Icon).count()
        if icons == 0:
            logger.info("Downloading icons...")
            try:
                r = requests.get(
                    "https://github.com/brian7704/OpenTAKServer-Installer/raw/master/iconsets.sqlite",
                    stream=True,
                )
                with open(
                    os.path.join(app.config.get("OTS_DATA_FOLDER"), "icons.sqlite"), "wb"
                ) as f:
                    f.write(r.content)

                def dict_factory(cursor, row):
                    d = {}
                    for idx, col in enumerate(cursor.description):
                        d[col[0]] = row[idx]
                    return d

                con = sqlite3.connect(
                    os.path.join(app.config.get("OTS_DATA_FOLDER"), "icons.sqlite")
                )
                con.row_factory = dict_factory
                cur = con.cursor()
                rows = cur.execute("SELECT * FROM icons")
                for row in rows:
                    db.session.execute(insert(Icon).values(**row))
                db.session.commit()
            except BaseException as e:
                logger.error("Failed to download icons: {}".format(e))
                logger.debug(traceback.format_exc())

        if app.config.get("DEBUG"):
            logger.debug("Starting in debug mode")
        else:
            logger.info("Starting in production mode")

        app.security.datastore.find_or_create_role(
            name="user", permissions={"user-read", "user-write"}
        )

        app.security.datastore.find_or_create_role(
            name="administrator", permissions={"administrator"}
        )

        # Make sure at least one admin user exists
        admin_user = db.session.execute(
            db.session.query(Role)
            .join(fsqla_v3.FsModels.roles_users)
            .where(Role.name == "administrator")
        ).scalar()
        if not admin_user:
            logger.info("Creating administrator account. The password is 'password'")
            app.security.datastore.create_user(
                username="administrator",
                password=hash_password("password"),
                roles=["administrator"],
            )
        db.session.commit()

    if app.config.get("OTS_ENABLE_MESHTASTIC"):
        mestastic_thread = MeshtasticController(app.app_context())
        app.mestastic_thread = mestastic_thread
    else:
        app.meshtastic_thread = None

    if app.config.get("OTS_ENABLE_MUMBLE_AUTHENTICATION"):
        try:
            logger.info("Starting Mumble authentication handler")
            mumble_daemon = MumbleIceDaemon(app, logger)
            mumble_daemon.daemon = True
            mumble_daemon.start()
        except BaseException as e:
            logger.error("Failed to enable Mumble authentication: {}".format(e))
            logger.error(traceback.format_exc())
    else:
        logger.info("Mumble authentication handler disabled")

    if app.config.get("OTS_ENABLE_PLUGINS"):
        try:
            app.plugin_manager = PluginManager(Plugin.group, app)
            app.plugin_manager.load_plugins()
            app.plugin_manager.activate(app)
        except BaseException as e:
            logger.error(f"Failed to load plugins: {e}")
            logger.debug(traceback.format_exc())

    app.start_time = datetime.now(timezone.utc)

    create_default_groups(app)

    try:
        socketio.run(
            app,
            host=app.config.get("OTS_LISTENER_ADDRESS"),
            port=app.config.get("OTS_LISTENER_PORT"),
            debug=app.config.get("DEBUG"),
            log_output=app.config.get("DEBUG"),
            use_reloader=False,
        )
    except KeyboardInterrupt:
        logger.warning("Caught CTRL+C, exiting...")
        if app.config.get("OTS_ENABLE_PLUGINS"):
            app.plugin_manager.stop_plugins()


def start():
    app = create_app(cli=False)

    @user_registered.connect_via(app)
    def user_registered_sighandler(app, user, confirmation_token, **kwargs):
        default_role = app.security.datastore.find_or_create_role(
            name="user", permissions={"user-read", "user-write"}
        )
        app.security.datastore.add_role_to_user(user, default_role)

    main(app)



"""Added some cascades

Revision ID: 2b5bddd207ae
Revises: 298430e7849d
Create Date: 2024-11-07 14:34:56.132740

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2b5bddd207ae"
down_revision = "298430e7849d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("alerts", schema=None) as batch_op:
            batch_op.drop_constraint("alert_cot", type_="foreignkey")
    except:
        pass
    try:
        with op.batch_alter_table("alerts", schema=None) as batch_op:
            batch_op.drop_constraint("alert_point", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("alerts", schema=None) as batch_op:
            batch_op.drop_constraint("alert_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("alerts", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "alert_point", "points", ["point_id"], ["id"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key(
            "alert_eud", "euds", ["sender_uid"], ["uid"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key("alert_cot", "cot", ["cot_id"], ["id"], ondelete="CASCADE")

    try:
        with op.batch_alter_table("casevac", schema=None) as batch_op:
            batch_op.drop_constraint("casevac_cot", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("casevac", schema=None) as batch_op:
            batch_op.drop_constraint("casevac_point", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("casevac", schema=None) as batch_op:
            batch_op.drop_constraint("casevac_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("casevac", schema=None) as batch_op:
        batch_op.create_foreign_key("casevac_cot", "cot", ["cot_id"], ["id"], ondelete="CASCADE")
        batch_op.create_foreign_key(
            "casevac_eud", "euds", ["sender_uid"], ["uid"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key(
            "casevac_point", "points", ["point_id"], ["id"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("certificates", schema=None) as batch_op:
            batch_op.drop_constraint("certificate_eud", type_="foreignkey")
    except BaseException as e:
        print(f"Shit failed {e}")

    with op.batch_alter_table("certificates", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "certificate_eud", "euds", ["eud_uid"], ["uid"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("markers", schema=None) as batch_op:
            batch_op.drop_constraint("marker_cot", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("markers", schema=None) as batch_op:
            batch_op.drop_constraint("marker_point", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("markers", schema=None) as batch_op:
            batch_op.drop_constraint("marker_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("markers", schema=None) as batch_op:
        batch_op.create_foreign_key("markers_cot", "cot", ["cot_id"], ["id"], ondelete="CASCADE")
        batch_op.create_foreign_key(
            "markers_eud", "euds", ["parent_uid"], ["uid"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key(
            "markers_point", "points", ["point_id"], ["id"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("points", schema=None) as batch_op:
            batch_op.drop_constraint("point_cot", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("points", schema=None) as batch_op:
            batch_op.drop_constraint("point_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("points", schema=None) as batch_op:
        batch_op.create_foreign_key("point_cot", "cot", ["cot_id"], ["id"], ondelete="CASCADE")
        batch_op.create_foreign_key(
            "point_eud", "euds", ["device_uid"], ["uid"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("rb_lines", schema=None) as batch_op:
            batch_op.drop_constraint("rb_line_cot", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("rb_lines", schema=None) as batch_op:
            batch_op.drop_constraint("rb_line_point", type_="foreignkey")
    except:
        pass

    try:
        with op.batch_alter_table("rb_lines", schema=None) as batch_op:
            batch_op.drop_constraint("rb_line_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("rb_lines", schema=None) as batch_op:
        batch_op.create_foreign_key("rb_line_cot", "cot", ["cot_id"], ["id"], ondelete="CASCADE")
        batch_op.create_foreign_key(
            "rb_line_eud", "euds", ["sender_uid"], ["uid"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key(
            "rb_line_point", "points", ["point_id"], ["id"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("zmist", schema=None) as batch_op:
            batch_op.drop_constraint("zmist_casevac", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("zmist", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "zmist_casevac", "casevac", ["casevac_uid"], ["uid"], ondelete="CASCADE"
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("zmist", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "casevac", ["casevac_uid"], ["uid"])

    with op.batch_alter_table("rb_lines", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "euds", ["sender_uid"], ["uid"])
        batch_op.create_foreign_key(None, "cot", ["cot_id"], ["id"])
        batch_op.create_foreign_key(None, "points", ["point_id"], ["id"])

    with op.batch_alter_table("points", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "cot", ["cot_id"], ["id"])
        batch_op.create_foreign_key(None, "euds", ["device_uid"], ["uid"])

    with op.batch_alter_table("markers", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "cot", ["cot_id"], ["id"])
        batch_op.create_foreign_key(None, "points", ["point_id"], ["id"])
        batch_op.create_foreign_key(None, "euds", ["parent_uid"], ["uid"])

    with op.batch_alter_table("certificates", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "euds", ["eud_uid"], ["uid"])

    with op.batch_alter_table("casevac", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "cot", ["cot_id"], ["id"])
        batch_op.create_foreign_key(None, "points", ["point_id"], ["id"])
        batch_op.create_foreign_key(None, "euds", ["sender_uid"], ["uid"])

    with op.batch_alter_table("apscheduler_jobs", schema=None) as batch_op:
        batch_op.alter_column(
            "id",
            existing_type=sa.String(length=255),
            type_=sa.VARCHAR(length=191),
            existing_nullable=False,
        )

    with op.batch_alter_table("alerts", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key(None, "points", ["point_id"], ["id"])
        batch_op.create_foreign_key(None, "cot", ["cot_id"], ["id"])
        batch_op.create_foreign_key(None, "euds", ["sender_uid"], ["uid"])

    # ### end Alembic commands ###



"""Added id column to device_profiles

Revision ID: 591a98184047
Revises: 61c7cfea0c86
Create Date: 2026-01-30 03:54:18.307669

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Sequence
from sqlalchemy.sql.ddl import CreateSequence

# revision identifiers, used by Alembic.
revision = "591a98184047"
down_revision = "61c7cfea0c86"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        try:
            batch_op.drop_constraint("device_profiles_pkey", type_="primary")
        except:
            pass
        batch_op.create_index(
            "preference_key_eud_uid",
            ["preference_key", "eud_uid"],
            unique=True,
            postgresql_nulls_not_distinct=False,
        )
        batch_op.execute(CreateSequence(Sequence("device_profiles_id_seq")))
        batch_op.add_column(
            sa.Column(
                "id",
                sa.Integer(),
                nullable=False,
                server_default=sa.text("nextval('device_profiles_id_seq'::regclass)"),
                autoincrement=True,
            )
        )
        batch_op.create_primary_key("primary_key", ["id"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.drop_constraint("primary_key", type_="primary")
        batch_op.create_primary_key("device_profiles_pkey", ["preference_key"])

    # ### end Alembic commands ###



"""Added mission_name to cot table

Revision ID: d31bff4a15c7
Revises: 5d06227dea50
Create Date: 2024-10-09 15:40:49.385642

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d31bff4a15c7"
down_revision = "5d06227dea50"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.add_column(sa.Column("mission_name", sa.String(length=255), nullable=True))
        batch_op.create_foreign_key("mission_name", "missions", ["mission_name"], ["name"])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.drop_constraint("mission_name", type_="foreignkey")
        batch_op.drop_column("mission_name")

    # ### end Alembic commands ###



import json
import math
import re
from datetime import datetime, timezone
from xml.etree.ElementTree import Element, SubElement, tostring

import pika.channel
from flask import current_app as app

ISO8601_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
ISO8601_FORMAT_NO_MICROSECONDS = "%Y-%m-%dT%H:%M:%SZ"
affiliations = [
    "friendly",
    "hostile",
    "unknown",
    "pending",
    "assumed",
    "neutral",
    "suspect",
    "joker",
    "faker",
]

# For WTForms BooleanField, the default doesn't include 'False'
# https://wtforms.readthedocs.io/en/3.1.x/fields/?highlight=false_values#wtforms.fields.BooleanField
false_values = (False, "False", "false", "")


def get_tasking(cot_type):
    if re.match("^t-x-f", cot_type):
        return "remarks"
    if re.match("^t-x-s", cot_type):
        return "state/sync"
    if re.match("^t-s", cot_type):
        return "required"
    if re.match("^t-z", cot_type):
        return "cancel"
    if re.match("^t-x-c-c", cot_type):
        return "commcheck"
    if re.match("^t-x-c-g-d", cot_type):
        return "dgps"
    if re.match("^t-k-d", cot_type):
        return "destroy"
    if re.match("^t-k-i", cot_type):
        return "investigate"
    if re.match("^t-k-t", cot_type):
        return "target"
    if re.match("^t-k", cot_type):
        return "strike"
    if re.match("^t-", cot_type):
        return "tasking"
    return None


def get_affiliation(cot_type):
    if re.match("^t-", cot_type):
        return get_tasking(cot_type)
    if re.match("^a-f-", cot_type):
        return "friendly"
    if re.match("^a-h-", cot_type):
        return "hostile"
    if re.match("^a-u-", cot_type):
        return "unknown"
    if re.match("^a-p-", cot_type):
        return "pending"
    if re.match("^a-a-", cot_type):
        return "assumed"
    if re.match("^a-n-", cot_type):
        return "neutral"
    if re.match("^a-s-", cot_type):
        return "suspect"
    if re.match("^a-j-", cot_type):
        return "joker"
    if re.match("^a-k-", cot_type):
        return "faker"
    return None


def get_battle_dimension(cot_type):
    if re.match("^a-.-A", cot_type):
        return "airborne"
    if re.match("^a-.-G", cot_type):
        return "ground"
    if re.match("^a-.-G-I", cot_type):
        return "installation"
    if re.match("^a-.-S", cot_type):
        return "surface/sea"
    if re.match("^a-.-U", cot_type):
        return "subsurface"
    if re.match("^a-.-F", cot_type):
        return "sof"
    if re.match("^a-.-Z", cot_type):
        return "unknown"
    if re.match("^a-.-P", cot_type):
        return "space"
    return None


def parse_type(cot_type):
    if re.match("^a-.-G-I", cot_type):
        return "installation"
    if re.match("^a-.-G-E-V", cot_type):
        return "vehicle"
    if re.match("^a-.-G-E", cot_type):
        return "equipment"
    if re.match("^a-.-A-W-M-S", cot_type):
        return "sam"
    if re.match("^a-.-A-M-F-Q-r", cot_type):
        return "uav"


def cot_type_to_2525c(cot_type):
    mil_std_2525c = "s"
    cot_type_list = cot_type.split("-")
    cot_type_list.pop(0)  # this should always be letter a
    affiliation = cot_type_list.pop(0)
    battle_dimension = cot_type_list.pop(0)
    mil_std_2525c += affiliation
    mil_std_2525c += battle_dimension
    mil_std_2525c += "-"

    for letter in cot_type_list:
        if letter.isupper():
            mil_std_2525c += letter.lower()

    while len(mil_std_2525c) < 10:
        mil_std_2525c += "-"

    return mil_std_2525c


def datetime_from_iso8601_string(datetime_string: str | None) -> datetime:
    if not datetime_string:
        return datetime.now(timezone.utc)
    try:
        dt = datetime.strptime(datetime_string, ISO8601_FORMAT)
        dt = dt.replace(tzinfo=timezone.utc)
        return dt
    except ValueError:
        dt = datetime.strptime(datetime_string, ISO8601_FORMAT_NO_MICROSECONDS)
        dt = dt.replace(tzinfo=timezone.utc)
        return dt


def iso8601_string_from_datetime(datetime_object):
    if datetime_object:
        return datetime_object.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-2] + "Z"
    else:
        return None


def iso8601_string_from_datetime_no_ms(datetime_object):
    if datetime_object:
        return datetime_object.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        return None


def generate_delete_cot(uid: str, cot_type: str) -> Element:
    now = datetime.now(timezone.utc)

    event = Element(
        "event",
        {
            "how": "h-g-i-g-o",
            "type": "t-x-d-d",
            "version": "2.0",
            "uid": uid,
            "start": iso8601_string_from_datetime(now),
            "time": iso8601_string_from_datetime(now),
            "stale": iso8601_string_from_datetime(now),
        },
    )
    SubElement(
        event, "point", {"ce": "9999999", "le": "9999999", "hae": "0", "lat": "0", "lon": "0"}
    )
    detail = SubElement(event, "detail")
    SubElement(detail, "link", {"relation": "p-p", "uid": uid, "type": cot_type})
    SubElement(
        detail,
        "_flow-tags_",
        {"TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0": iso8601_string_from_datetime(now)},
    )

    return event


def format_bytes(size_bytes: int | None):
    if size_bytes is None or size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"


def bytes_to_megabytes(size_bytes: int | None):
    if not size_bytes:
        return 0
    return round(size_bytes / 1024**2, 2)


def bytes_to_gigabytes(size_bytes: int | None):
    if not size_bytes:
        return 0
    return round(size_bytes / 1024**3, 2)



from dataclasses import dataclass
from datetime import datetime

from sqlalchemy import BigInteger, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.Chatrooms import Chatroom

# Leave these imports, they're needed when making new DB migrations
from opentakserver.models.Team import Team


@dataclass
class EUD(db.Model):
    __tablename__ = "euds"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    device: Mapped[str] = mapped_column(String(255), nullable=True)
    os: Mapped[str] = mapped_column(String(255), nullable=True)
    platform: Mapped[str] = mapped_column(String(255), nullable=True)
    version: Mapped[str] = mapped_column(String(255), nullable=True)
    phone_number: Mapped[int] = mapped_column(BigInteger, nullable=True)
    last_event_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    last_status: Mapped[str] = mapped_column(String(255), nullable=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey("user.id"), nullable=True)
    team_id: Mapped[int] = mapped_column(Integer, ForeignKey("teams.id"), nullable=True)
    team_role: Mapped[str] = mapped_column(String(255), nullable=True)
    meshtastic_id: Mapped[int] = mapped_column(BigInteger, nullable=True)
    meshtastic_macaddr: Mapped[String] = mapped_column(String(255), nullable=True)
    last_meshtastic_publish: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    points = relationship("Point", cascade="all, delete-orphan", back_populates="eud")
    cots = relationship("CoT", cascade="all, delete-orphan", back_populates="eud")
    casevacs = relationship("CasEvac", cascade="all, delete-orphan", back_populates="eud")
    geochats = relationship("GeoChat", cascade="all, delete-orphan", back_populates="eud")
    chatroom_uid = relationship("ChatroomsUids", cascade="all, delete-orphan", back_populates="eud")
    user = relationship("User", back_populates="euds")
    alert = relationship("Alert", cascade="all, delete-orphan", back_populates="eud")
    data_packages = relationship(
        "DataPackage", cascade="all, delete-orphan", back_populates="eud", uselist=False
    )
    certificate = relationship(
        "Certificate", cascade="all, delete, delete-orphan", back_populates="eud", uselist=False
    )
    markers = relationship("Marker", cascade="all, delete, delete-orphan", back_populates="eud")
    rb_lines = relationship("RBLine", cascade="all, delete-orphan", back_populates="eud")
    team = relationship("Team", back_populates="euds")
    owned_missions = relationship("Mission", back_populates="owner")
    stats = relationship("EUDStats", back_populates="eud")
    profiles = relationship("DeviceProfiles", back_populates="eud")

    def serialize(self):
        return {
            "uid": self.uid,
            "callsign": self.callsign,
            "device": self.device,
            "os": self.os,
            "platform": self.platform,
            "version": self.version,
            "phone_number": self.phone_number,
            "last_event_time": self.last_event_time,
            "last_status": self.last_status,
            "user_id": self.user_id,
            "team_id": self.team_id,
            "team_role": self.team_role,
        }

    def to_json(self, include_data_packages=True):
        config_datapackage_hash = None
        if self.certificate and self.certificate.data_package:
            config_datapackage_hash = self.certificate.data_package.hash
        return {
            "uid": self.uid,
            "callsign": self.callsign,
            "device": self.device,
            "os": self.os,
            "platform": self.platform,
            "version": self.version,
            "phone_number": self.phone_number,
            "last_event_time": (
                iso8601_string_from_datetime(self.last_event_time) if self.last_event_time else None
            ),
            "last_status": self.last_status,
            "username": self.user.username if self.user else None,
            "last_point": None,  # Setting to None for now since it can cause a huge overhead when an EUD has lots of points in the DB
            "team": self.team.name if self.team else None,
            "team_color": self.team.get_team_color() if self.team else None,
            "team_role": self.team_role,
            "data_packages": (
                self.data_packages.to_json(False)
                if include_data_packages and self.data_packages
                else None
            ),
            "config_hash": config_datapackage_hash,
        }



import datetime
import enum
from dataclasses import dataclass

from sqlalchemy import Boolean, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db


class InvitationTypeEnum(str, enum.Enum):
    clientUid = "clientUid"
    callsign = "callsign"
    userName = "userName"
    group = "group"
    team = "team"


@dataclass
class MissionInvitation(db.Model):
    __tablename__ = "mission_invitations"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"))
    client_uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True
    )
    callsign: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.callsign", ondelete="CASCADE"), nullable=True
    )
    username: Mapped[str] = mapped_column(String(255), ForeignKey("user.username"), nullable=True)
    group_name: Mapped[str] = mapped_column(String(255), nullable=True)
    team_name: Mapped[str] = mapped_column(String(255), ForeignKey("teams.name"), nullable=True)
    creator_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    role: Mapped[str] = mapped_column(String(255), nullable=True)
    type: Mapped[str] = mapped_column(String(255), nullable=True, default="callsign")

    eud_uid = relationship("EUD", foreign_keys=[client_uid], uselist=False)
    eud_callsign = relationship("EUD", foreign_keys=[callsign], uselist=False)
    user = relationship("User", back_populates="mission_invitations", uselist=False)
    team = relationship("Team", back_populates="mission_invitations", uselist=False)
    mission = relationship("Mission", back_populates="invitations", uselist=False)

    def serialize(self):
        return {
            "mission_name": self.mission_name,
            "client_uid": self.client_uid,
            "callsign": self.callsign,
            "username": self.username,
            "group_name": self.group,
            "team_name": self.team_name,
            "creator_uid": self.creator_uid,
            "role": self.role,
            "type": self.type,
        }

    def to_json(self):
        return self.serialize()



import base64
from dataclasses import dataclass

from sqlalchemy import INTEGER, Integer, LargeBinary, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db


@dataclass
class Icon(db.Model):
    __tablename__ = "icons"

    id: Mapped[int] = mapped_column(INTEGER, primary_key=True)
    iconset_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    filename: Mapped[str] = mapped_column(String(255), nullable=True)
    groupName: Mapped[str] = mapped_column(String(255), nullable=True)
    type2525b: Mapped[str] = mapped_column(String(255), nullable=True)
    useCnt: Mapped[int] = mapped_column(Integer, nullable=True)
    bitmap: Mapped[bytes] = mapped_column(LargeBinary, nullable=True)
    shadow: Mapped[bytes] = mapped_column(LargeBinary, nullable=True)
    markers = relationship("Marker", back_populates="icon")

    def serialize(self):
        return {
            "iconset_uid": self.iconset_uid,
            "filename": self.filename,
            "groupName": self.groupName,
            "type2525b": self.type2525b,
            "useCnt": self.useCnt,
            "bitmap": (
                "data:image/png;base64,{}".format(base64.b64encode(self.bitmap).decode("utf-8"))
                if self.bitmap
                else None
            ),
            "shadow": (
                "data:image/png;base64,{}".format(base64.b64encode(self.shadow).decode("utf-8"))
                if self.shadow
                else None
            ),
        }

    def to_json(self):
        return self.serialize()


class IconSets(db.Model):
    __tablename__ = "iconsets"

    id: Mapped[int] = mapped_column(INTEGER, primary_key=True)
    name: Mapped[str] = mapped_column(String(255), nullable=True)
    uid: Mapped[str] = mapped_column(String(255), nullable=True)
    version: Mapped[int] = mapped_column(Integer, nullable=True)
    defaultFriendly: Mapped[str] = mapped_column(String(255), nullable=True)
    defaultHostile: Mapped[str] = mapped_column(String(255), nullable=True)
    defaultNeutral: Mapped[str] = mapped_column(String(255), nullable=True)
    defaultUnknown: Mapped[str] = mapped_column(String(255), nullable=True)
    selectedGroup: Mapped[str] = mapped_column(String(255), nullable=True)



from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db


class ChatroomsUids(db.Model):
    __tablename__ = "chatrooms_uids"

    chatroom_id: Mapped[str] = mapped_column(
        String(255), ForeignKey("chatrooms.id"), primary_key=True
    )
    uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), primary_key=True
    )
    chatroom = relationship("Chatroom", back_populates="chatroom_uid")
    eud = relationship("EUD", back_populates="chatroom_uid")



import math
from dataclasses import dataclass
from datetime import datetime

from pygc import great_circle
from sqlalchemy import Boolean, DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class RBLine(db.Model):
    __tablename__ = "rb_lines"

    # type = u-rb-a
    # how = h-e

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    sender_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"))
    uid: Mapped[str] = mapped_column(String(255), unique=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)

    range: Mapped[float] = mapped_column(Float)
    bearing: Mapped[float] = mapped_column(Float)
    inclination: Mapped[float] = mapped_column(Float, nullable=True)
    anchor_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    range_units: Mapped[int] = mapped_column(Integer, nullable=True)
    bearing_units: Mapped[int] = mapped_column(Integer, nullable=True)
    north_ref: Mapped[int] = mapped_column(Integer, nullable=True)
    color: Mapped[int] = mapped_column(Integer, nullable=True)
    color_hex: Mapped[str] = mapped_column(String(255), nullable=True)
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    stroke_color: Mapped[int] = mapped_column(Integer, nullable=True)
    stroke_weight: Mapped[float] = mapped_column(Float, nullable=True)
    stroke_style: Mapped[str] = mapped_column(String(255), nullable=True)
    labels_on: Mapped[bool] = mapped_column(Boolean, nullable=True)
    point_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("points.id", ondelete="CASCADE"), nullable=True
    )
    cot_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("cot.id", ondelete="CASCADE"), nullable=True
    )
    end_latitude: Mapped[float] = mapped_column(Float, nullable=True)
    end_longitude: Mapped[float] = mapped_column(Float, nullable=True)
    point = relationship("Point", back_populates="rb_line")
    cot = relationship("CoT", back_populates="rb_line")
    eud = relationship("EUD", back_populates="rb_lines")

    range_unit_names = ["standard", "metric", "nautical"]
    bearing_unit_names = ["degrees", "mils"]
    north_ref_names = ["true", "magnetic", "grid"]

    def color_to_hex(self):
        if self.color:
            return format(int(self.color) & 0xFFFFFFFF, "08X")

    def calc_end_point(self, start_point):
        if not int(self.bearing_units):
            azimuth = float(self.bearing)
        else:
            azimuth = math.degrees(float(self.bearing))

        return great_circle(
            distance=float(self.range),
            azimuth=azimuth,
            latitude=start_point.latitude,
            longitude=start_point.longitude,
        )

    def serialize(self):
        return {
            "uid": self.uid,
            "timestamp": self.timestamp,
            "range": self.range,
            "inclination": self.inclination,
            "anchor_uid": self.anchor_uid,
            "range_units": self.range_units,
            "bearing_units": self.bearing_units,
            "north_ref": self.north_ref,
            "color": self.color,
            "color_hex": self.color_hex,
            "stroke_weight": self.stroke_weight,
            "stroke_style": self.stroke_style,
            "labels_on": self.labels_on,
            "end_latitude": self.end_latitude,
            "end_longitude": self.end_longitude,
        }

    def to_json(self):
        return {
            "uid": self.uid,
            "timestamp": iso8601_string_from_datetime(self.timestamp),
            "range": self.range,
            "inclination": self.inclination,
            "anchor_uid": self.anchor_uid,
            "range_units": self.range_units,
            "range_unit_name": (
                self.range_unit_names[int(self.range_units)] if self.range_units else None
            ),
            "bearing_units": self.bearing_units,
            "bearing_unit_name": (
                self.bearing_unit_names[int(self.bearing_units)] if self.bearing else None
            ),
            "north_ref": self.north_ref,
            "north_ref_name": self.north_ref_names[int(self.north_ref)] if self.north_ref else None,
            "color": self.color,
            "color_hex": self.color_hex,
            "stroke_weight": self.stroke_weight,
            "stroke_style": self.stroke_style,
            "labels_on": self.labels_on,
            "point": self.point.to_json() if self.point else None,
            "end_latitude": self.end_latitude,
            "end_longitude": self.end_longitude,
        }



"""Updated MissionLogEntry

Revision ID: 35a075531973
Revises: 2b1863a985a6
Create Date: 2025-11-12 05:01:51.094540

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "35a075531973"
down_revision = "2b1863a985a6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("mission_logs", schema=None) as batch_op:
        batch_op.add_column(sa.Column("content_hashes", sa.JSON(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_logs", schema=None) as batch_op:
        batch_op.drop_column("content_hashes")

    # ### end Alembic commands ###



"""More cascades

Revision ID: 795ebb9262d8
Revises: 2b5bddd207ae
Create Date: 2024-11-07 19:14:25.263371

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "795ebb9262d8"
down_revision = "2b5bddd207ae"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("certificates", schema=None) as batch_op:
            batch_op.drop_constraint("certificate_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("certificates", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "certificate_eud", "euds", ["eud_uid"], ["uid"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("chatrooms_uids", schema=None) as batch_op:
            batch_op.drop_constraint("chatroom_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("chatrooms_uids", schema=None) as batch_op:
        batch_op.create_foreign_key("chatroom_eud", "euds", ["uid"], ["uid"], ondelete="CASCADE")

    try:
        with op.batch_alter_table("cot", schema=None) as batch_op:
            batch_op.drop_constraint("cot_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.create_foreign_key("cot_eud", "euds", ["sender_uid"], ["uid"], ondelete="CASCADE")

    try:
        with op.batch_alter_table("data_packages", schema=None) as batch_op:
            batch_op.drop_constraint("data_packages_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("data_packages", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "data_packages_eud", "euds", ["creator_uid"], ["uid"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("geochat", schema=None) as batch_op:
            batch_op.drop_constraint("geochat_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("geochat", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "geochat_eud", "euds", ["sender_uid"], ["uid"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("groups_euds", schema=None) as batch_op:
            batch_op.drop_constraint("group_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("groups_euds", schema=None) as batch_op:
        batch_op.create_foreign_key("group_eud", "euds", ["eud_uid"], ["uid"], ondelete="CASCADE")

    try:
        with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
            batch_op.drop_constraint("mission_invitation_callsign", type_="foreignkey")
            batch_op.drop_constraint("mission_invitation_eud", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "mission_invitation_eud", "euds", ["client_uid"], ["uid"], ondelete="CASCADE"
        )
        batch_op.create_foreign_key(
            "mission_invitation_callsign", "euds", ["callsign"], ["callsign"], ondelete="CASCADE"
        )

    try:
        with op.batch_alter_table("missions", schema=None) as batch_op:
            batch_op.drop_constraint("mission_creator", type_="foreignkey")
    except:
        pass

    with op.batch_alter_table("missions", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "mission_creator", "euds", ["creator_uid"], ["uid"], ondelete="CASCADE"
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("missions", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("mission_creator", "euds", ["creator_uid"], ["uid"])

    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("mission_invitation_eud", "euds", ["client_uid"], ["uid"])
        batch_op.create_foreign_key(
            "mission_invitation_callsign", "euds", ["callsign"], ["callsign"]
        )

    with op.batch_alter_table("groups_euds", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("group_eud", "euds", ["eud_uid"], ["uid"])

    with op.batch_alter_table("geochat", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("geochat_eud", "euds", ["sender_uid"], ["uid"])

    with op.batch_alter_table("data_packages", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("data_packages_eud", "euds", ["creator_uid"], ["uid"])

    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("cot_eud", "euds", ["sender_uid"], ["uid"])

    with op.batch_alter_table("chatrooms_uids", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("chatroom_eud", "euds", ["uid"], ["uid"])

    with op.batch_alter_table("certificates", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.create_foreign_key("certificate_eud", "euds", ["eud_uid"], ["uid"])

    # ### end Alembic commands ###



from datetime import timedelta, timezone
from uuid import UUID

import pika
import sqlalchemy.exc
from flask import Blueprint, jsonify, request
from flask_babel import gettext
from flask_security import auth_required, current_user
from sqlalchemy import insert, update
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.blueprints.ots_api.api import paginate, route_cot, search
from opentakserver.extensions import db, logger, socketio
from opentakserver.forms.casevac_form import CasEvacForm
from opentakserver.forms.zmist_form import ZmistForm
from opentakserver.functions import *
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.CoT import CoT
from opentakserver.models.EUD import EUD
from opentakserver.models.Point import Point
from opentakserver.models.ZMIST import ZMIST

casevac_api_blueprint = Blueprint("casevac_api_blueprint", __name__)


@casevac_api_blueprint.route("/api/casevac", methods=["GET"])
@auth_required()
def query_casevac():
    query = db.session.query(CasEvac)

    query = search(query, EUD, "callsign")
    query = search(query, CasEvac, "sender_uid")
    query = search(query, CasEvac, "uid")

    return paginate(query, CasEvac)


@casevac_api_blueprint.route("/api/casevac", methods=["POST"])
@auth_required()
def add_casevac():
    form = CasEvacForm(formdata=ImmutableMultiDict(request.json))
    if not form.validate():
        return jsonify({"success": False, "errors": form.errors}), 400

    zmist = None
    if "zmist" in request.json.keys():
        logger.warning(request.json["zmist"])
        zmist_form = ZmistForm(formdata=ImmutableMultiDict(request.json["zmist"]))
        if not zmist_form.validate():
            return jsonify({"success": False, "errors": form.errors}), 400
        zmist = ZMIST()
        zmist.from_wtform(zmist_form)

    casevac = CasEvac()
    casevac.from_wtforms(form)
    casevac.sender_uid = app.config.get("OTS_NODE_ID")
    if zmist:
        casevac.zmist = zmist

    point = Point()
    point.from_wtform(form)
    point.device_uid = app.config.get("OTS_NODE_ID")

    casevac.point = point

    cot = CoT()
    cot.how = "h-g-i-g-o"
    cot.type = "b-r-f-h-c"
    cot.sender_uid = app.config.get("OTS_NODE_ID")
    cot.sender_callsign = app.config.get("OTS_NODE_ID")
    cot.timestamp = point.timestamp
    cot.start = point.timestamp
    cot.stale = point.timestamp + timedelta(days=365)
    cot.xml = tostring(casevac.to_cot()).decode("utf-8")

    cot_result = db.session.execute(insert(CoT).values(**cot.serialize()))
    db.session.commit()

    cot_pk = cot_result.inserted_primary_key[0]
    point.cot_id = cot_pk
    casevac.cot_id = cot_pk

    point_result = db.session.execute(insert(Point).values(**point.serialize()))
    db.session.commit()
    point_pk = point_result.inserted_primary_key[0]

    casevac.point_id = point_pk

    try:
        db.session.add(casevac)

        if zmist:
            zmist.casevac_uid = casevac.uid
            db.session.add(zmist)
        db.session.commit()
        logger.debug(f"Saved new CasEvac {casevac.uid}")

    except sqlalchemy.exc.IntegrityError:
        db.session.rollback()
        db.session.execute(
            update(CasEvac)
            .where(CasEvac.uid == casevac.uid)
            .values(point_id=point_pk, cot_id=casevac.cot_id, **casevac.serialize())
        )
        db.session.commit()
        logger.debug(f"Updated CasEvac {casevac.uid}")

    rabbit_credentials = pika.PlainCredentials(
        app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD")
    )
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(
        pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials)
    )
    channel = rabbit_connection.channel()
    channel.basic_publish(
        exchange="cot",
        routing_key="",
        body=json.dumps({"cot": cot.xml, "uid": app.config["OTS_NODE_ID"]}),
        properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")),
    )
    route_cot(cot.xml, current_user)
    channel.close()
    rabbit_connection.close()

    casevac.zmist = zmist
    socketio.emit("casevac", casevac.to_json(), namespace="/socket.io")

    return jsonify({"success": True}), 200


@casevac_api_blueprint.route("/api/casevac", methods=["DELETE"])
@auth_required()
def delete_casevac():
    uid = request.args.get("uid")
    if not uid:
        return jsonify({"success": False, "error": "Please specify a UID"}), 400

    try:
        UUID(uid, version=4)
    except ValueError:
        return jsonify({"success": False, "error": f"Invalid UID: {uid}"}), 400

    query = db.session.query(CasEvac)
    query = search(query, CasEvac, "uid")
    casevac = db.session.execute(query).first()
    if not casevac:
        return jsonify({"success": False, "error": gettext("Unknown UID: %(uid)s", uid=uid)}), 404

    casevac = casevac[0]

    now = datetime.now(timezone.utc)
    event = Element(
        "event",
        {
            "how": "h-g-i-g-o",
            "type": "t-x-d-d",
            "version": "2.0",
            "uid": casevac.uid,
            "start": iso8601_string_from_datetime(now),
            "time": iso8601_string_from_datetime(now),
            "stale": iso8601_string_from_datetime(now),
        },
    )
    SubElement(
        event, "point", {"ce": "9999999", "le": "9999999", "hae": "0", "lat": "0", "lon": "0"}
    )
    detail = SubElement(event, "detail")
    SubElement(detail, "link", {"relation": "p-p", "uid": casevac.uid, "type": casevac.cot.type})
    SubElement(
        detail,
        "_flow-tags_",
        {"TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0": iso8601_string_from_datetime(now)},
    )

    rabbit_credentials = pika.PlainCredentials(
        app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD")
    )
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(
        pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials)
    )
    channel = rabbit_connection.channel()
    channel.basic_publish(
        exchange="cot",
        routing_key="",
        body=json.dumps({"cot": tostring(event).decode("utf-8"), "uid": app.config["OTS_NODE_ID"]}),
        properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")),
    )
    route_cot(tostring(event).decode("utf-8"), current_user)
    channel.close()
    rabbit_connection.close()

    db.session.delete(casevac)
    db.session.commit()

    return jsonify({"success": True})



"""Added groups_missions table

Revision ID: 2b1863a985a6
Revises: eaf2af6f6b22
Create Date: 2025-11-09 18:33:51.885010

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2b1863a985a6"
down_revision = "eaf2af6f6b22"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "groups_missions",
        sa.Column("mission_name", sa.String(length=255), nullable=False),
        sa.Column("group_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["groups.id"],
        ),
        sa.ForeignKeyConstraint(
            ["mission_name"],
            ["missions.name"],
        ),
        sa.PrimaryKeyConstraint("mission_name", "group_id"),
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_table("groups_missions")
    # ### end Alembic commands ###



from dataclasses import dataclass
from datetime import datetime

from sqlalchemy import DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class Certificate(db.Model):
    __tablename__ = "certificates"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    common_name: Mapped[str] = mapped_column(String(255))
    eud_uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True
    )
    data_package_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("data_packages.id"), nullable=True
    )
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    username: Mapped[str] = mapped_column(String(255), ForeignKey("user.username"), nullable=True)
    expiration_date: Mapped[datetime] = mapped_column(DateTime)
    server_address: Mapped[str] = mapped_column(String(255))
    server_port: Mapped[int] = mapped_column(Integer)
    truststore_filename: Mapped[str] = mapped_column(String(255))
    user_cert_filename: Mapped[str] = mapped_column(String(255))
    csr: Mapped[str] = mapped_column(String(255), nullable=True)
    cert_password: Mapped[str] = mapped_column(String(255))
    user = relationship("User", back_populates="certificate", uselist=False)
    eud = relationship("EUD", cascade="all, delete", back_populates="certificate", uselist=False)
    data_package = relationship("DataPackage", back_populates="certificate", uselist=False)

    def serialize(self):
        return {
            "callsign": self.callsign,
            "expiration_date": self.expiration_date,
            "server_address": self.server_address,
            "server_port": self.server_port,
            "truststore_filename": self.truststore_filename,
            "user_cert_filename": self.user_cert_filename,
            "cert_password": self.cert_password,
            "eud_uid": self.eud_uid,
        }

    def to_json(self):
        return {
            "callsign": self.callsign,
            "expiration_date": iso8601_string_from_datetime(self.expiration_date),
            "server_address": self.server_address,
            "server_port": self.server_port,
            "truststore_filename": self.truststore_filename,
            "user_cert_filename": self.user_cert_filename,
            "data_package_filename": self.data_package.filename if self.data_package else None,
            "data_package_hash": self.data_package.hash if self.data_package else None,
            "eud_uid": self.eud_uid,
        }



"""Added columns to mission_invtations table

Revision ID: f6dfc571d31c
Revises: 14e268184cc5
Create Date: 2024-10-17 18:53:11.141218

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f6dfc571d31c"
down_revision = "14e268184cc5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.add_column(sa.Column("creator_uid", sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column("role", sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.drop_column("role")
        batch_op.drop_column("creator_uid")

    # ### end Alembic commands ###



"""Made cot.sender_callsign nullable

Revision ID: 85dbe0a7dcc1
Revises: 6665d030f187
Create Date: 2025-01-30 15:19:53.796637

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "85dbe0a7dcc1"
down_revision = "6665d030f187"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.alter_column(
            "sender_callsign", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.alter_column(
            "sender_callsign", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    # ### end Alembic commands ###



import os
import shutil
import traceback
from datetime import datetime, timezone

from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request, send_from_directory
from flask_babel import gettext
from flask_security import auth_required
from sqlalchemy import update
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.blueprints.marti_api.data_package_marti_api import data_package_share
from opentakserver.blueprints.ots_api.api import paginate, search
from opentakserver.extensions import db, logger
from opentakserver.forms.data_package_form import DataPackageUpdateForm
from opentakserver.models.Certificate import Certificate
from opentakserver.models.DataPackage import DataPackage

data_package_api = Blueprint("data_package_api", __name__)


@data_package_api.route("/api/data_packages", methods=["PATCH"])
@auth_required()
def edit_data_package():
    form = DataPackageUpdateForm(formdata=ImmutableMultiDict(request.json))
    if not form.validate():
        return jsonify({"success": False, "errors": form.errors}), 400

    data_package = db.session.execute(
        db.session.query(DataPackage).filter_by(hash=form.hash.data)
    ).first()
    if not data_package:
        return (
            jsonify({"success": False, "error": f"Package with hash {form.hash.data} not found"}),
            404,
        )

    data_package = data_package[0]

    if data_package.filename.endswith("_CONFIG.zip"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "Server connection data packages can't be installed on enrollment or connection"
                    ),
                }
            ),
            400,
        )

    if form.install_on_enrollment.data is not None:
        data_package.install_on_enrollment = form.install_on_enrollment.data
    if form.install_on_connection.data is not None:
        data_package.install_on_connection = form.install_on_connection.data

    data_package.submission_time = datetime.now(timezone.utc)

    db.session.execute(
        update(DataPackage)
        .filter(DataPackage.hash == data_package.hash)
        .values(**data_package.serialize())
    )
    db.session.commit()

    return jsonify({"success": True})


@data_package_api.route("/api/data_packages", methods=["DELETE"])
@auth_required()
def delete_data_package():
    file_hash = request.args.get("hash")
    if not file_hash:
        return jsonify({"success": False, "error": "Please provide a file hash"}), 400

    query = db.session.query(DataPackage)
    query = search(query, DataPackage, "hash")
    data_package = db.session.execute(query).first()
    if not data_package:
        return jsonify({"success": False, "error": gettext("Invalid/unknown hash")}), 400

    try:
        logger.warning(
            "Deleting data package {} - {}".format(data_package[0].filename, data_package[0].hash)
        )
        db.session.delete(data_package[0])
        db.session.commit()
        os.remove(
            os.path.join(app.config.get("UPLOAD_FOLDER"), "{}.zip".format(data_package[0].hash))
        )

        if data_package[0].certificate:
            Certificate.query.filter_by(id=data_package[0].certificate.id).delete()
            db.session.commit()
            shutil.rmtree(
                os.path.join(
                    app.config.get("OTS_CA_FOLDER"),
                    "certs",
                    data_package[0].certificate.common_name,
                ),
                ignore_errors=True,
            )
    except BaseException as e:
        logger.error("Failed to delete data package")
        logger.error(traceback.format_exc())
        return jsonify({"success": False, "error": str(e)}), 500

    return jsonify({"success": True})


@data_package_api.route("/api/data_packages")
@auth_required()
def data_packages():
    query = db.session.query(DataPackage)
    query = search(query, DataPackage, "filename")
    query = search(query, DataPackage, "hash")
    query = search(query, DataPackage, "creator_uid")
    query = search(query, DataPackage, "keywords")
    query = search(query, DataPackage, "mime_type")
    query = search(query, DataPackage, "size")
    query = search(query, DataPackage, "tool")

    return paginate(query, DataPackage)


@data_package_api.route("/api/data_packages/download")
@auth_required()
def data_package_download():
    if "hash" not in request.args.keys():
        return (
            jsonify({"success": False, "error": gettext("Please provide a data package hash")}),
            400,
        )

    file_hash = request.args.get("hash")

    query = db.session.query(DataPackage)
    query = search(query, DataPackage, "hash")

    data_package = db.session.execute(query).first()

    if not data_package:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("Data package with hash '%(hash)s' not found", hash=file_hash),
                }
            ),
            404,
        )

    download_name = data_package[0].filename
    name, extension = os.path.splitext(download_name)

    return send_from_directory(
        app.config.get("UPLOAD_FOLDER"),
        f"{file_hash}{extension}",
        as_attachment=True,
        download_name=download_name,
    )


@data_package_api.route("/api/data_packages", methods=["POST"])
@auth_required()
def upload_data_package():
    return data_package_share()



"""More data sync stuff

Revision ID: d1f5df78eace
Revises: 4f0173cdb93b
Create Date: 2024-10-15 03:02:43.059600

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d1f5df78eace"
down_revision = "4f0173cdb93b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_changes", schema=None) as batch_op:
        batch_op.add_column(sa.Column("content_uid", sa.String(length=255), nullable=True))
        batch_op.create_foreign_key("content_resource", "mission_content", ["content_uid"], ["uid"])

    with op.batch_alter_table("mission_content", schema=None) as batch_op:
        batch_op.add_column(sa.Column("filename", sa.String(length=255), nullable=True))
        batch_op.drop_column("name")

    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.add_column(sa.Column("callsign", sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column("username", sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column("group_name", sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column("team_name", sa.String(length=255), nullable=True))
        batch_op.alter_column("client_uid", existing_type=sa.String(length=255), nullable=True)
        batch_op.create_foreign_key("mission_invitation_user", "user", ["username"], ["username"])
        # batch_op.create_foreign_key('groups', 'groups', ['group_name'], ['group_name'])
        batch_op.create_foreign_key(
            "mission_invitation_callsign", "euds", ["callsign"], ["callsign"]
        )
        batch_op.create_foreign_key("mission_invitation_team", "teams", ["team_name"], ["name"])
        batch_op.create_foreign_key("mission_invitation_eud", "euds", ["client_uid"], ["uid"])

    with op.batch_alter_table("missions", schema=None) as batch_op:
        batch_op.add_column(sa.Column("group", sa.String(length=255), nullable=True))
        # batch_op.create_foreign_key('groups', 'groups', ['group'], ['group_name'])
        batch_op.drop_column("group_name")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("missions", schema=None) as batch_op:
        batch_op.add_column(sa.Column("group_name", sa.String(length=255), nullable=True))
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_column("group")

    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.drop_constraint("user", type_="foreignkey")
        batch_op.drop_constraint("eud_callsign", type_="foreignkey")
        batch_op.drop_constraint("eud_uid", type_="foreignkey")
        batch_op.drop_constraint("teams", type_="foreignkey")
        batch_op.alter_column("client_uid", existing_type=sa.String(length=255), nullable=False)
        batch_op.drop_column("team_name")
        batch_op.drop_column("group_name")
        batch_op.drop_column("username")
        batch_op.drop_column("callsign")

    with op.batch_alter_table("mission_content", schema=None) as batch_op:
        batch_op.add_column(sa.Column("name", sa.String(length=255), nullable=True))
        batch_op.drop_column("filename")

    with op.batch_alter_table("mission_changes", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_column("content_uid")

    # ### end Alembic commands ###



import argparse
import logging
import os
import platform
import sys
from logging.handlers import TimedRotatingFileHandler

import colorlog
import flask_wtf
import sqlalchemy
import yaml
from flask import Flask, jsonify
from flask_security import Security, SQLAlchemyUserDatastore
from flask_security.models import fsqla

from opentakserver.defaultconfig import DefaultConfig
from opentakserver.EmailValidator import EmailValidator

# These unused imports are required by SQLAlchemy, don't remove them
from opentakserver.eud_handler.SocketServer import SocketServer
from opentakserver.extensions import db, ldap_manager, logger
from opentakserver.models.Alert import Alert
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.Certificate import Certificate
from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.CoT import CoT
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.DeviceProfiles import DeviceProfiles
from opentakserver.models.EUD import EUD
from opentakserver.models.EUDStats import EUDStats
from opentakserver.models.Group import Group
from opentakserver.models.GroupMission import GroupMission
from opentakserver.models.Marker import Marker
from opentakserver.models.Mission import Mission
from opentakserver.models.MissionChange import MissionChange
from opentakserver.models.MissionContentMission import MissionContentMission
from opentakserver.models.MissionInvitation import MissionInvitation
from opentakserver.models.MissionLogEntry import MissionLogEntry
from opentakserver.models.MissionUID import MissionUID
from opentakserver.models.Point import Point
from opentakserver.models.RBLine import RBLine
from opentakserver.models.Team import Team
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.WebAuthn import WebAuthn
from opentakserver.models.ZMIST import ZMIST
from opentakserver.PasswordValidator import PasswordValidator


def args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ssl", help="Enable SSL", default=False, action=argparse.BooleanOptionalAction
    )
    return parser.parse_args()


def setup_logging(app):
    level = logging.INFO
    if app.config.get("DEBUG"):
        level = logging.DEBUG
    logger.setLevel(level)

    if sys.stdout.isatty():
        color_log_handler = colorlog.StreamHandler()
        color_log_formatter = colorlog.ColoredFormatter(
            "%(log_color)s[%(asctime)s] - eud_handler[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S %Z",
        )
        color_log_handler.setFormatter(color_log_formatter)
        logger.addHandler(color_log_handler)

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs"), exist_ok=True)
    fh = TimedRotatingFileHandler(
        os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs", "opentakserver.log"),
        when=app.config.get("OTS_LOG_ROTATE_WHEN"),
        interval=app.config.get("OTS_LOG_ROTATE_INTERVAL"),
        backupCount=app.config.get("OTS_BACKUP_COUNT"),
    )
    fh.setFormatter(
        logging.Formatter(
            "[%(asctime)s] - eud_handler[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s"
        )
    )
    logger.addHandler(fh)


def create_app():
    app = Flask(__name__)
    app.config.from_object(DefaultConfig)

    # Load config.yml if it exists
    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
        app.config.from_file(
            os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load
        )
    else:
        # First run, created config.yml based on default settings
        logger.info("Creating config.yml")
        with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
            conf = {}
            for option in DefaultConfig.__dict__:
                # Fix the sqlite DB path on Windows
                if (
                    option == "SQLALCHEMY_DATABASE_URI"
                    and platform.system() == "Windows"
                    and DefaultConfig.__dict__[option].startswith("sqlite")
                ):
                    conf[option] = (
                        DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
                    )
                elif option.isupper():
                    conf[option] = DefaultConfig.__dict__[option]
            config.write(yaml.safe_dump(conf))

    setup_logging(app)
    db.init_app(app)

    if app.config.get("OTS_ENABLE_LDAP"):
        logger.info("Enabling LDAP")
        ldap_manager.init_app(app)

    # The rest is required by flask, leave it in
    try:
        fsqla.FsModels.set_db_info(db)
    except sqlalchemy.exc.InvalidRequestError:
        pass

    from opentakserver.models.role import Role
    from opentakserver.models.user import User

    flask_wtf.CSRFProtect(app)
    user_datastore = SQLAlchemyUserDatastore(db, User, Role)
    app.security = Security(
        app, user_datastore, mail_util_cls=EmailValidator, password_util_cls=PasswordValidator
    )

    return app


app = create_app()


@app.route("/status")
def status():
    return jsonify({"status": "ok"})


def main():
    opts = args()
    if opts.ssl:
        socket_server = SocketServer(
            logger, app.app_context(), app.config.get("OTS_SSL_STREAMING_PORT"), True
        )
        logger.info(f"Started SSL server on port {app.config.get('OTS_SSL_STREAMING_PORT')}")
    else:
        socket_server = SocketServer(
            logger, app.app_context(), app.config.get("OTS_TCP_STREAMING_PORT")
        )
        logger.info(f"Started TCP server on port {app.config.get('OTS_TCP_STREAMING_PORT')}")
    socket_server.run()


if __name__ == "__main__":
    main()



from dataclasses import dataclass

from flask_security.models import fsqla_v3 as fsqla

from opentakserver.extensions import db


@dataclass
class Role(db.Model, fsqla.FsRoleMixin):

    def serialize(self):
        return {
            "name": self.name,
            "description": self.description,
            "permissions": self.permissions,
            "update_timestamp": self.update_datetime,
        }

    def __eq__(self, other):
        return self.name == other or self.name == getattr(other, "name", None)



import traceback

import bleach
import sqlalchemy
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request
from flask_babel import gettext
from flask_security import (
    admin_change_password,
    auth_required,
    current_user,
    hash_password,
    roles_accepted,
)

from opentakserver.blueprints.ots_api.api import paginate, search
from opentakserver.extensions import db, ldap_manager, logger
from opentakserver.models.EUD import EUD
from opentakserver.models.Group import Group
from opentakserver.models.GroupUser import GroupUser
from opentakserver.models.user import User
from opentakserver.UsernameValidator import UsernameValidator

user_api_blueprint = Blueprint("user_api_blueprint", __name__)


@user_api_blueprint.route("/api/user/add", methods=["POST"])
@roles_accepted("administrator")
def create_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to create users"
                    ),
                }
            ),
            400,
        )

    username = bleach.clean(request.json.get("username"))
    password = bleach.clean(request.json.get("password"))
    confirm_password = bleach.clean(request.json.get("confirm_password"))

    validated_username = UsernameValidator(app).validate(username)
    if validated_username[0]:
        return jsonify({"success": False, "error": f"{validated_username[0]}"}), 400

    if password != confirm_password:
        return jsonify({"success": False, "error": gettext("Passwords do not match")}), 400

    roles = request.json.get("roles")
    roles_cleaned = []

    for role in roles:
        role = bleach.clean(role)
        role_exists = app.security.datastore.find_role(role)

        if not role_exists:
            return (
                jsonify(
                    {"success": False, "error": gettext("Role %(role)s does not exist", role=role)}
                ),
                400,
            )

        elif role == "administrator" and not current_user.has_role("administrator"):
            return (
                jsonify(
                    {
                        "success": False,
                        "error": gettext(
                            "Only administrators can add users to the administrators role"
                        ),
                    }
                ),
                403,
            )

        elif role not in roles_cleaned:
            roles_cleaned.append(role)

    if not app.security.datastore.find_user(username=username):
        logger.info("Creating user {}".format(username))
        app.security.datastore.create_user(
            username=username, password=hash_password(password), roles=roles_cleaned
        )
        db.session.commit()
        return jsonify({"success": True}), 200
    else:
        logger.error("User {} already exists".format(username))
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s already exists", username=username),
                }
            ),
            400,
        )


@user_api_blueprint.route("/api/user/delete", methods=["POST"])
@roles_accepted("administrator")
def delete_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to delete users"
                    ),
                }
            ),
            400,
        )

    username = bleach.clean(request.json.get("username"))

    if username == current_user.username:
        return (
            jsonify({"success": False, "error": gettext("You can't delete your own account")}),
            400,
        )

    logger.info("Deleting user {}".format(username))

    try:
        user = app.security.datastore.find_user(username=username)
        db.session.execute(sqlalchemy.delete(GroupUser).where(GroupUser.user_id == user.id))
        app.security.datastore.delete_user(user)
    except BaseException as e:
        logger.error(traceback.format_exc())
        return (
            jsonify({"success": False, "error": gettext("Failed to delete user: %(e)s", e=str(e))}),
            400,
        )

    db.session.commit()
    return jsonify({"success": True}), 200


@user_api_blueprint.route("/api/user/password/reset", methods=["POST"])
@roles_accepted("administrator")
def admin_reset_password():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to reset passwords"
                    ),
                }
            ),
            400,
        )

    username = bleach.clean(request.json.get("username"))
    new_password = bleach.clean(request.json.get("new_password"))

    if not username or not new_password:
        return (
            jsonify(
                {"success": False, "error": gettext("Please specify a username and new password")}
            ),
            400,
        )

    if len(new_password) < app.config.get("SECURITY_PASSWORD_LENGTH_MIN"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "Your password must be at least %(characters)s characters long",
                        characters=app.config.get("SECURITY_PASSWORD_LENGTH_MIN"),
                    ),
                }
            ),
            400,
        )

    user = app.security.datastore.find_user(username=username)
    if user:
        admin_change_password(user, new_password, False)
        db.session.commit()
        return jsonify({"success": True}), 200
    else:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("Could not find user %(username)s", username=username),
                }
            ),
            400,
        )


@user_api_blueprint.route("/api/user/deactivate", methods=["POST"])
@roles_accepted("administrator")
def deactivate_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to deactivate users"
                    ),
                }
            ),
            400,
        )

    username = bleach.clean(request.json.get("username", ""))
    if not username:
        return (
            jsonify(
                {"success": False, "error": gettext("Please specify the username to deactivate")}
            ),
            400,
        )

    user = app.security.datastore.find_user(username=username)
    if not user:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s does not exist", username=username),
                }
            ),
            400,
        )

    deactivated = app.security.datastore.deactivate_user(user)
    if deactivated:
        db.session.commit()
        return jsonify({"success": True})
    else:
        return jsonify(
            {
                "success": False,
                "error": gettext("%(username)s is already deactivated", username=username),
            }
        )


@user_api_blueprint.route("/api/user/activate", methods=["POST"])
@roles_accepted("administrator")
def activate_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to activate users"
                    ),
                }
            ),
            400,
        )

    username = bleach.clean(request.json.get("username", ""))
    if not username:
        return (
            jsonify(
                {"success": False, "error": gettext("Please specify the username to activate")}
            ),
            400,
        )

    user = app.security.datastore.find_user(username=username)
    if not user:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s does not exist", username=username),
                }
            ),
            400,
        )

    activated = app.security.datastore.activate_user(user)
    if activated:
        db.session.commit()
        return jsonify({"success": True})
    else:
        return jsonify(
            {
                "success": False,
                "error": gettext("%(username)s is already activated", username=username),
            }
        )


@user_api_blueprint.route("/api/user/role", methods=["POST"])
@roles_accepted("administrator")
def set_user_role():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to assign roles"
                    ),
                }
            ),
            400,
        )

    username = bleach.clean(request.json.get("username", ""))
    roles = request.json.get("roles")
    roles_cleaned = []

    if not username or not roles:
        return (
            jsonify({"success": False, "error": gettext("Please specify a username and roles")}),
            400,
        )

    for role in roles:
        role = bleach.clean(role)
        role_exists = app.security.datastore.find_role(role)

        if not role_exists:
            return (
                jsonify(
                    {"success": False, "error": gettext("Role %(role)s does not exist", role=role)}
                ),
                400,
            )

        elif role not in roles_cleaned:
            roles_cleaned.append(role)

    user = app.security.datastore.find_user(username=username)
    if not user:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s does not exist", username=username),
                }
            ),
            400,
        )

    for role in user.roles:
        app.security.datastore.remove_role_from_user(user, role)

    for role in roles_cleaned:
        app.security.datastore.add_role_to_user(user, role)

    db.session.commit()
    return jsonify({"success": True})


@user_api_blueprint.route("/api/user/assign_eud", methods=["POST"])
@auth_required()
def assign_eud_to_user():
    username = bleach.clean(request.json.get("username")) if "username" in request.json else None
    eud_uid = bleach.clean(request.json.get("uid")) if "uid" in request.json else None
    user = None

    if not eud_uid:
        return (
            {"success": False, "error": "Please specify an EUD"},
            400,
            {"Content-Type": "application/json"},
        )
    if not username or username == current_user.username:
        user = current_user
    elif username != current_user.username and current_user.has_role("administrator"):
        user = app.security.datastore.find_user(username=username)
        if not user:
            return (
                jsonify(
                    {
                        "success": False,
                        "error": gettext("User %(username)s does not exist", username=username),
                    }
                ),
                404,
            )

    eud = db.session.query(EUD).filter_by(uid=eud_uid).first()

    if not eud:
        return (
            jsonify(
                {"success": False, "error": gettext("EUD %(eud_uid)s not found", eud_uid=eud_uid)}
            ),
            404,
        )
    elif (
        eud.user_id
        and not current_user.has_role("administrator")
        and current_user.id != eud.user_id
    ):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("%(uid)s is already assigned to another user", uid=eud.uid),
                }
            ),
            403,
        )
    else:
        eud.user_id = user.id
        db.session.add(eud)
        db.session.commit()

        return jsonify({"success": True})


@user_api_blueprint.route("/api/users")
@roles_accepted("administrator")
def get_users():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to manage users"
                    ),
                }
            ),
            400,
        )

    query = db.session.query(User)
    query = search(query, User, "username")

    return paginate(query)


@user_api_blueprint.route("/api/users/all")
@roles_accepted("administrator")
def get_all_users():
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to manage users"
                    ),
                }
            ),
            400,
        )

    users = db.session.execute(db.session.query(User)).all()
    return_value = []

    for user in users:
        user = user[0]
        return_value.append(user.serialize())

    return return_value


@user_api_blueprint.route("/api/users/groups")
@roles_accepted("administrator")
def get_user_groups():
    """Gets a list of group memberships for a user
    :parameter: username

    :return: List of group memberships
    """
    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to manage groups"
                    ),
                }
            ),
            400,
        )

    username = request.args.get("username")
    if not username:
        return jsonify({"success": False, "error": gettext("Please provide a username")}), 400

    username = bleach.clean(username)

    user = app.security.datastore.find_user(username=username)
    if not user:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s not found", username=username),
                }
            ),
            404,
        )

    group_memberships = db.session.execute(
        db.session.query(GroupUser).filter_by(user_id=user.id)
    ).all()
    memberships = []
    for membership in group_memberships:
        membership: GroupUser = membership[0]
        memberships.append(
            {
                "group_name": membership.group.name,
                "direction": membership.direction,
                "active": membership.enabled,
            }
        )

    return jsonify({"success": True, "results": memberships})


@user_api_blueprint.route("/api/users/groups", methods=["PUT"])
@roles_accepted("administrator")
def add_user_to_groups():
    """Adds a user to one or more groups
    :parameter: groups - List of groups to add a user to
    :parameter: username
    :parameter: direction - Group direction, must be either IN or OUT

    :return: 400 if LDAP is enabled, no group or username is specified, or if the specified group or user doesn't exist or the user is already in the group. 200 on success.
    :rtype: Response
    """

    if app.config.get("OTS_ENABLE_LDAP"):
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "LDAP is enabled, please use your LDAP server to add users to groups"
                    ),
                }
            ),
            400,
        )

    groups = request.json.get("groups")
    username = request.json.get("username")
    direction = request.json.get("direction")

    if not groups or not username or not direction:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "Please provide a list of groups, a username, and a direction"
                    ),
                }
            ),
            400,
        )

    if direction != "IN" and direction != "OUT":
        return jsonify({"success": False, "error": gettext("Direction must be IN or OUT")}), 400

    user = app.security.datastore.find_user(username=username)
    if not user:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s doesn't exist", username=username),
                }
            ),
            404,
        )

    for group_name in groups:
        group_name = bleach.clean(group_name)
        group = db.session.execute(db.session.query(Group).filter_by(name=group_name)).first()
        if not group:
            return (
                jsonify(
                    {
                        "success": False,
                        "error": gettext(
                            "Group %(group_name)s doesn't exist", group_name=group_name
                        ),
                    }
                ),
                404,
            )

        group = group[0]

        membership = GroupUser()
        membership.user_id = user.id
        membership.group_id = group.id
        membership.direction = direction

        try:
            db.session.add(membership)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()

    return jsonify({"success": True})



"""Set install on enrollment and connection to nullable

Revision ID: 6af2256c568d
Revises: 34d5df698b81
Create Date: 2024-09-20 20:17:28.926142

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6af2256c568d"
down_revision = "34d5df698b81"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("data_packages", schema=None) as batch_op:
        batch_op.alter_column("install_on_enrollment", existing_type=sa.BOOLEAN(), nullable=True)
        batch_op.alter_column("install_on_connection", existing_type=sa.BOOLEAN(), nullable=True)

    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.alter_column("enrollment", existing_type=sa.BOOLEAN(), nullable=True)
        batch_op.alter_column("connection", existing_type=sa.BOOLEAN(), nullable=True)
        try:
            batch_op.drop_column("id")
        except BaseException as e:
            print(f"Failed to drop euds.id column: {e}")
            pass

    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.alter_column("install_on_enrollment", existing_type=sa.BOOLEAN(), nullable=True)
        batch_op.alter_column("install_on_connection", existing_type=sa.BOOLEAN(), nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.alter_column("install_on_connection", existing_type=sa.BOOLEAN(), nullable=False)
        batch_op.alter_column("install_on_enrollment", existing_type=sa.BOOLEAN(), nullable=False)

    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.add_column(sa.Column("id", sa.INTEGER(), nullable=False))
        batch_op.alter_column("connection", existing_type=sa.BOOLEAN(), nullable=False)
        batch_op.alter_column("enrollment", existing_type=sa.BOOLEAN(), nullable=False)

    with op.batch_alter_table("data_packages", schema=None) as batch_op:
        batch_op.alter_column("install_on_connection", existing_type=sa.BOOLEAN(), nullable=False)
        batch_op.alter_column("install_on_enrollment", existing_type=sa.BOOLEAN(), nullable=False)

    # ### end Alembic commands ###



import datetime
import json
import os
import pathlib
import traceback
import uuid
from urllib.parse import urlparse

import bleach
import requests
import sqlalchemy.exc
from ffmpeg import FFmpeg
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request
from flask_babel import gettext
from flask_ldap3_login import AuthenticationResponseStatus
from flask_security import auth_required, current_user, verify_password
from flask_security.utils import parse_auth_token
from sqlalchemy import update
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.extensions import db, ldap_manager, logger
from opentakserver.forms.MediaMTXPathConfig import MediaMTXPathConfig
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.VideoStream import VideoStream

mediamtx_api_blueprint = Blueprint("mediamtx_api_blueprint", __name__)


def get_stream_protocol(source_type):
    protocol = "rtsp"
    if source_type.startswith("rtsps"):
        protocol = "rtsps"
    if source_type.startswith("rtsp"):
        protocol = "rtsp"
    elif source_type == "hlsSource":
        protocol = "hls"
    elif source_type == "rpiCameraSource":
        protocol = "rpi_camera"
    elif source_type.startswith("rtmp"):
        protocol = "rtmp"
    elif source_type.startswith("srt"):
        protocol = "srt"
    elif source_type.startswith("udp"):
        protocol = "udp"
    elif source_type.startswith("webRTC"):
        protocol = "webrtc"

    return protocol


@mediamtx_api_blueprint.route("/api/mediamtx/webhook")
def mediamtx_webhook():
    token = request.args.get("token")
    if not token or bleach.clean(token) != app.config.get("OTS_MEDIAMTX_TOKEN"):
        logger.error("Invalid token")
        return jsonify({"success": False, "error": gettext("Invalid token")}), 401

    event = bleach.clean(request.args.get("event"))
    if event == "init":
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
        path = bleach.clean(request.args.get("path"))

        if path == "startup":
            paths = VideoStream.query.all()
            for path in paths:
                r = requests.post(
                    "{}/v3/config/paths/add/{}".format(
                        app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path.path
                    ),
                    json=json.loads(path.mediamtx_settings),
                )
                logger.debug("Init added {} {}".format(path, r.status_code))

            # Get all paths from MediaMTX and make sure they're in OTS's database
            r = requests.get("{}/v3/paths/list".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS")))
            paths = r.json()
            for path in paths["items"]:
                video_stream = (
                    db.session.query(VideoStream).where(VideoStream.path == path["name"]).first()
                )
                if not video_stream:
                    if not path["source"]:
                        continue
                    video_stream = VideoStream()
                    video_stream.protocol = get_stream_protocol(path["source"]["type"])

                    r = requests.get(
                        "{}/v3/config/global/get".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"))
                    )
                    video_stream.port = r.json()["rtspAddress"].replace(":", "")

                    r = requests.get(
                        "{}/v3/config/paths/get/{}".format(
                            app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path["name"]
                        )
                    )
                    video_stream.mediamtx_settings = json.dumps(r.json())

                    video_stream.path = path["name"]
                    video_stream.alias = path["name"]
                    video_stream.rtsp_reliable = 1
                    video_stream.ready = event == "ready"
                    video_stream.rover_port = -1
                    video_stream.ignore_embedded_klv = False
                    video_stream.buffer_time = None
                    video_stream.network_timeout = 10000
                    video_stream.uid = str(uuid.uuid4())
                    video_stream.generate_xml(urlparse(request.url_root).hostname)

                    db.session.add(video_stream)
                    db.session.commit()

    elif event == "connect":
        connection_type = bleach.clean(request.args.get("connection_type"))
        connection_id = bleach.clean(request.args.get("connection_id"))
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
    elif event == "ready" or event == "notready":
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
        path = bleach.clean(request.args.get("path"))
        query = bleach.clean(request.args.get("query"))
        source_type = bleach.clean(request.args.get("source_type"))
        source_id = bleach.clean(request.args.get("source_id"))

        video_stream = db.session.query(VideoStream).where(VideoStream.path == path).first()
        if video_stream:
            video_stream.ready = event == "ready"
            db.session.add(video_stream)
            db.session.commit()
            r = requests.patch(
                "{}/v3/config/paths/patch/{}".format(
                    app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path
                ),
                json=json.loads(video_stream.mediamtx_settings),
            )
            logger.debug("Ready Patched path {}: {} - {}".format(path, r.status_code, r.text))
        else:
            video_stream = VideoStream()
            if source_type.startswith("rtsps"):
                video_stream.protocol = "rtsps"
            if source_type.startswith("rtsp"):
                video_stream.protocol = "rtsp"
            elif source_type == "hlsSource":
                video_stream.protocol = "hls"
            elif source_type == "rpiCameraSource":
                video_stream.protocol = "rpi_camera"
            elif source_type.startswith("rtmp"):
                video_stream.protocol = "rtmp"
            elif source_type.startswith("srt"):
                video_stream.protocol = "srt"
            elif source_type.startswith("udp"):
                video_stream.protocol = "udp"
            elif source_type.startswith("webRTC"):
                video_stream.protocol = "webrtc"

            # video_stream.query = query
            video_stream.port = rtsp_port
            video_stream.path = path
            video_stream.alias = path
            video_stream.rtsp_reliable = 1
            video_stream.ready = event == "ready"
            video_stream.rover_port = -1
            video_stream.ignore_embedded_klv = False
            video_stream.buffer_time = None
            video_stream.network_timeout = 10000
            video_stream.uid = str(uuid.uuid4())
            video_stream.generate_xml(urlparse(request.url_root).hostname)
            mediamtx_settings = MediaMTXPathConfig(None)
            mediamtx_settings.sourceOnDemand.data = source_id is not None
            mediamtx_settings.record.data = False
            video_stream.mediamtx_settings = json.dumps(mediamtx_settings.serialize())

            db.session.add(video_stream)
            db.session.commit()

        if event == "ready":
            os.makedirs(
                os.path.join(
                    app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", video_stream.path
                ),
                exist_ok=True,
            )

            try:
                (
                    FFmpeg()
                    .input(video_stream.to_json()["rtsp_link"] + "?token={}".format(token))
                    .option("y")
                    .output(
                        os.path.join(
                            app.config.get("OTS_DATA_FOLDER"),
                            "mediamtx",
                            "recordings",
                            video_stream.path,
                            "thumbnail.png",
                        ),
                        {"frames:v": 1},
                    )
                    .execute()
                )
            except BaseException as e:
                logger.error(f"Failed to create thumbnail: {e}")
                logger.debug(traceback.format_exc())

    elif event == "read":
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
        path = bleach.clean(request.args.get("path"))
        query = bleach.clean(request.args.get("query"))
        reader_type = bleach.clean(request.args.get("reader_type"))
        reader_id = bleach.clean(request.args.get("reader_id"))
    elif event == "disconnect":
        connection_type = bleach.clean(request.args.get("connection_type"))
        connection_id = bleach.clean(request.args.get("connection_id"))
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
    elif event == "segment_record":
        recording = VideoRecording()
        recording.segment_path = bleach.clean(request.args.get("segment_path"))
        recording.path = bleach.clean(request.args.get("path"))
        recording.in_progress = True
        recording.start_time = datetime.datetime.now(datetime.timezone.utc)

        with app.app_context():
            try:
                db.session.add(recording)
                db.session.commit()
            except sqlalchemy.exc.IntegrityError:
                db.session.rollback()
                db.session.execute(
                    update(VideoRecording)
                    .filter(VideoRecording.segment_path == recording.segment_path)
                    .values(**recording.serialize())
                )
                db.session.commit()
    elif event == "segment_record_complete":
        segment_path = bleach.clean(request.args.get("segment_path"))
        with app.app_context():
            recording = db.session.execute(
                db.session.query(VideoRecording).filter(VideoRecording.segment_path == segment_path)
            ).first()
            if recording and recording.count:
                recording = recording[0]
                recording.in_progress = False
                recording.stop_time = datetime.datetime.now(datetime.timezone.utc)
                recording.duration = (
                    recording.stop_time - recording.start_time.replace(tzinfo=datetime.timezone.utc)
                ).seconds
            else:
                recording = VideoRecording()
                recording.segment_path = bleach.clean(request.args.get("segment_path"))
                recording.path = bleach.clean(request.args.get("path"))
                recording.in_progress = False
                recording.stop_time = datetime.datetime.now(datetime.timezone.utc)

            try:
                probe = json.loads(
                    FFmpeg(executable="ffprobe")
                    .input(
                        recording.segment_path,
                        print_format="json",
                        show_streams=None,
                        show_format=None,
                    )
                    .execute()
                )
                for stream in probe["streams"]:
                    if stream["codec_type"].lower() == "video":
                        recording.width = stream["width"]
                        recording.height = stream["height"]
                        recording.video_bitrate = stream["bit_rate"]
                        recording.video_codec = stream["codec_name"]
                        FFmpeg().input(recording.segment_path, ss="00:00:01").option("y").output(
                            recording.segment_path + ".png", {"frames:v": 1}
                        ).execute()
                    elif stream["codec_type"].lower() == "audio":
                        recording.audio_codec = stream["codec_name"]
                        recording.audio_samplerate = stream["sample_rate"]
                        recording.audio_channels = stream["channels"]
                        recording.audio_bitrate = stream["bit_rate"]
                if "format" in probe and "size" in probe["format"]:
                    recording.file_size = probe["format"]["size"]
            except BaseException as e:
                logger.error(f"Failed to run ffprobe: {e}")
                logger.debug(traceback.format_exc())

            db.session.add(recording)
            db.session.commit()
        pass

    return "", 200


@mediamtx_api_blueprint.route("/api/mediamtx/stream/add", methods=["POST"])
@mediamtx_api_blueprint.route("/api/mediamtx/stream/update", methods=["PATCH"])
@auth_required()
def add_update_stream():
    try:
        form = MediaMTXPathConfig(formdata=ImmutableMultiDict(request.json))
        if not form.validate():
            return jsonify({"success": False, "errors": form.errors}), 400

        path = bleach.clean(request.json.get("path", ""))

        if not path:
            return jsonify({"success": False, "error": gettext("Please specify a path name")}), 400

        if path.startswith("/"):
            return (
                jsonify({"success": False, "error": gettext("Path cannot begin with a slash")}),
                400,
            )

        video = db.session.query(VideoStream).where(VideoStream.path == path).first()
        if not video and request.path.endswith("add"):
            video = VideoStream()
            video.path = path
            video.username = current_user.username
            video.mediamtx_settings = json.dumps(form.serialize())
            video.rover_port = -1
            video.ignore_embedded_klv = False
            video.buffer_time = None
            video.rtsp_reliable = 1
            video.network_timeout = 10000
            video.generate_xml(urlparse(request.url_root).hostname)
        elif not video and request.path.endswith("update"):
            return (
                jsonify({"success": False, "error": gettext("Path %(path)s not found", path=path)}),
                400,
            )

        settings = json.loads(video.mediamtx_settings)

        for setting in request.json:
            if setting == "csrf_token" or setting == "sourceOnDemand" or setting == "path":
                continue

            # When the source URL is blank set sourceOnDemand to false in order to avoid a MediaMTX error
            if setting == "source":
                source = request.json.get(setting)
                if not source:
                    settings["sourceOnDemand"] = False
                # Run ffmpeg and yt-dlp for live YouTube videos
                elif "youtube.com" in source or "youtu.be" in source:
                    settings["source"] = None
                    settings["sourceOnDemand"] = False
                    yt_dlp_path = os.path.join(
                        pathlib.Path.home(), ".opentakserver_venv", "bin", "yt-dlp"
                    )
                    settings["runOnDemand"] = (
                        f"sh -c 'ffmpeg -re -i \"$({yt_dlp_path} -g {source})\" -c:v copy -f rtsp rtsp://127.0.0.1:8554/$RTSP_PATH'"
                    )
                continue

            key = bleach.clean(setting)
            value = request.json.get(setting)
            if isinstance(value, str):
                value = bleach.clean(value)
            if value is not None:
                settings[key] = value
                logger.debug("set {} to {}".format(key, value))

        if request.path.endswith("update"):
            r = requests.patch(
                "{}/v3/config/paths/patch/{}".format(
                    app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path
                ),
                json=settings,
            )
        else:
            r = requests.post(
                "{}/v3/config/paths/add/{}".format(
                    app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path
                ),
                json=settings,
            )

        if r.status_code == 200:
            logger.debug("Patched path {}: {}".format(path, r.status_code))
            video.mediamtx_settings = json.dumps(settings)
            db.session.add(video)
            db.session.commit()
            return jsonify({"success": True})

        else:
            action = "add" if request.path.endswith("add") else "update"
            logger.error(
                "Failed to {} mediamtx path: {} - {}".format(
                    action, r.status_code, r.json()["error"]
                )
            )
            return jsonify({"success": False, "error": r.json()["error"]}), 400
    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({"success": False, "error": str(e)}), 500


@mediamtx_api_blueprint.route("/api/mediamtx/stream/delete", methods=["DELETE"])
@auth_required()
def delete_stream():
    try:
        path = bleach.clean(request.args.get("path", ""))

        if not path:
            return jsonify({"success": False, "error": gettext("Please specify a path name")}), 400

        r = requests.delete(
            "{}/v3/config/paths/delete/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path)
        )
        logger.debug("Delete status code: {}".format(r.status_code))
        video = db.session.query(VideoStream).filter(VideoStream.path == path)
        if not video:
            return (
                jsonify({"success": False, "error": gettext("Path %(path)s not found", path=path)}),
                400,
            )

        video.delete()
        db.session.commit()
    except requests.exceptions.ConnectionError as e:
        logger.error(traceback.format_exc())
        return jsonify({"success": False, "error": gettext("MediaMTX is not running")}), 500

    if r.status_code != 404:
        return r.text, r.status_code
    else:
        return "", 200


# This is mainly for mediamtx authentication
@mediamtx_api_blueprint.route("/api/external_auth", methods=["POST"])
def external_auth():
    username = bleach.clean(request.json.get("user"))
    password = bleach.clean(request.json.get("password"))
    action = bleach.clean(request.json.get("action"))
    query = bleach.clean(request.json.get("query"))
    ip = bleach.clean(request.json.get("ip"))

    # Whitelist 127.0.0.1 to make things like YouTube video re-streaming work
    if ip and ip in app.config.get("OTS_IP_WHITELIST"):
        return "", 200

    # Token auth to prevent high CPU usage when reading HLS streams
    if "jwt" in query or "token" in query:
        query = query.split("&")
        for q in query:
            if "=" not in q:
                continue
            key, value = q.split("=")
            if key == "jwt":
                try:
                    parse_auth_token(value)
                    return "", 200
                except BaseException as e:
                    logger.error(f"Invalid token: {e}")
                    return "", 401
            elif key == "token":
                if value == app.config.get("OTS_MEDIAMTX_TOKEN"):
                    return "", 200
                else:
                    return "", 401

    auth_success = False

    # LDAP Auth
    if app.config.get("OTS_ENABLE_LDAP"):
        result = ldap_manager.authenticate(username, password)
        if result.status == AuthenticationResponseStatus.success:
            # Keep this import here to avoid a circular import when OTS is started
            from opentakserver.blueprints.ots_api.ldap_api import save_user

            save_user(result.user_dn, result.user_id, result.user_info, result.user_groups)
            auth_success = True
        else:
            return "", 401
    # Flask-Security auth
    else:
        user = app.security.datastore.find_user(username=username)
        if not user:
            return "", 401
        if not verify_password(password, user.password):
            return "", 401
        auth_success = True

    if auth_success:
        if action == "publish":
            logger.debug("Publish {}".format(request.json.get("path")))
            v = VideoStream()
            v.uid = bleach.clean(request.json.get("id")) if request.json.get("id") else None
            v.rover_port = -1
            v.ignore_embedded_klv = False
            v.buffer_time = None
            v.network_timeout = 10000
            v.protocol = bleach.clean(request.json.get("protocol"))
            v.path = bleach.clean(request.json.get("path"))
            v.alias = v.path.split("/")[-1]
            v.username = bleach.clean(request.json.get("user"))
            path_config = MediaMTXPathConfig(None).serialize()
            path_config["sourceOnDemand"] = False
            v.mediamtx_settings = json.dumps(path_config)

            if v.protocol == "rtsp":
                v.port = 8554
                v.rtsp_reliable = 1
            elif v.protocol == "rtmp":
                v.port = 1935
                v.rtsp_reliable = 1
            else:
                v.rtsp_reliable = 0

            v.generate_xml(request.json.get("ip"))

            with app.app_context():
                try:

                    db.session.add(v)
                    db.session.commit()
                    r = requests.post(
                        "{}/v3/config/paths/add/{}".format(
                            app.config.get("OTS_MEDIAMTX_API_ADDRESS"), v.path
                        ),
                        json=path_config,
                    )
                    if r.status_code == 200:
                        logger.debug("Added path {} to mediamtx".format(v.path))
                    else:
                        logger.error(
                            "Failed to add path {} to mediamtx. Status code {} {}".format(
                                v.path, r.status_code, r.text
                            )
                        )
                    logger.debug("Inserted video stream {}".format(v.uid))
                except sqlalchemy.exc.IntegrityError as e:
                    try:
                        db.session.rollback()
                        video = (
                            db.session.query(VideoStream).filter(VideoStream.path == v.path).first()
                        )
                        r = requests.post(
                            "{}/v3/config/paths/add/{}".format(
                                app.config.get("OTS_MEDIAMTX_API_ADDRESS"), v.path
                            ),
                            json=json.loads(video.mediamtx_settings),
                        )
                        if r.status_code == 200:
                            logger.debug("Added path {} to mediamtx".format(v.path))
                        else:
                            logger.error(
                                "Failed to add path {} to mediamtx. Status code {} {}".format(
                                    v.path, r.status_code, r.text
                                )
                            )
                    except:
                        logger.error(traceback.format_exc())

        logger.debug("external_auth returning 200")
        return "", 200
    elif query:
        for arg in query.split("&"):
            key, value = arg.split("=")
            if key == "token" and value == app.config.get("OTS_MEDIAMTX_TOKEN"):
                return "", 200
    else:
        logger.debug("external_auth returning 401")
        return "", 401



import base64
import datetime
import os
import traceback
from urllib.parse import urlparse
from xml.etree.ElementTree import Element, SubElement, tostring

import bleach
import sqlalchemy
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request
from flask_ldap3_login import AuthenticationResponseStatus
from flask_security import verify_password
from OpenSSL import crypto

from opentakserver.certificate_authority import CertificateAuthority
from opentakserver.extensions import db, ldap_manager, logger
from opentakserver.models.Certificate import Certificate
from opentakserver.models.EUD import EUD
from opentakserver.models.Token import Token

certificate_authority_api_blueprint = Blueprint("certificate_authority_api_blueprint", __name__)


# flask-security's http_auth_required() decorator will deny access because ATAK doesn't do CSRF,
# so we handle basic auth ourselves
def basic_auth(credentials):
    try:
        username, password = (
            base64.b64decode(credentials.split(" ", 1)[-1].encode("utf-8"))
            .decode("utf-8")
            .split(":", 1)
        )
        username = bleach.clean(username)
        password = bleach.clean(password)

        if app.config.get("OTS_ENABLE_LDAP"):
            result = ldap_manager.authenticate(username, password)

            if result.status == AuthenticationResponseStatus.success:
                # Keep this import here to avoid a circular import when OTS is started
                from opentakserver.blueprints.ots_api.ldap_api import save_user

                save_user(result.user_dn, result.user_id, result.user_info, result.user_groups)
                return True
            return False

        user = app.security.datastore.find_user(username=username)
        if not user:
            logger.error(f"User {username} doesn't exist")
            return False
        elif verify_password(password, user.password):
            return True
        else:
            return Token.verify_token(password)

    except BaseException as e:
        logger.error("Failed to verify credentials: {}".format(e))
        return False


# require basic auth
@certificate_authority_api_blueprint.route("/Marti/api/tls/config")
def tls_config():
    root_element = Element("ns2:certificateConfig")
    root_element.set("xmlns", "http://bbn.com/marti/xml/config")
    root_element.set("xmlns:ns2", "com.bbn.marti.config")

    name_entries = SubElement(root_element, "nameEntries")
    first_name_entry = SubElement(name_entries, "nameEntry")
    first_name_entry.set("name", "O")
    first_name_entry.set("value", app.config.get("OTS_CA_ORGANIZATION"))

    second_name_entry = SubElement(name_entries, "nameEntry")
    second_name_entry.set("name", "OU")
    second_name_entry.set("value", app.config.get("OTS_CA_ORGANIZATIONAL_UNIT"))

    return tostring(root_element), 200, {"Content-Type": "application/xml"}


@certificate_authority_api_blueprint.route("/Marti/api/tls/signClient/", methods=["POST"])
def sign_csr():
    if not basic_auth(request.headers.get("Authorization")):
        return "", 401
    return "", 200


@certificate_authority_api_blueprint.route("/Marti/api/tls/signClient/v2", methods=["POST"])
def sign_csr_v2():
    if not basic_auth(request.headers.get("Authorization")):
        return "", 401

    try:
        if "clientUID" in request.args.keys():
            uid = request.args.get("clientUID")
        else:
            uid = request.args.get("clientUid")

        csr = request.data.decode("utf-8")
        if "BEGIN CERTIFICATE REQUEST" not in csr:
            csr = "-----BEGIN CERTIFICATE REQUEST-----\n" + csr

        if not csr.endswith("\n"):
            csr = csr + "\n"

        csr = csr + "-----END CERTIFICATE REQUEST-----"

        x509 = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr.encode())

        common_name = x509.get_subject().CN
        logger.debug("Attempting to sign CSR for {}".format(common_name))

        cert_authority = CertificateAuthority(logger, app)

        signed_csr = cert_authority.sign_csr(csr.encode(), common_name, False).decode("utf-8")
        signed_csr = signed_csr.replace("-----BEGIN CERTIFICATE-----\n", "")
        signed_csr = signed_csr.replace("\n-----END CERTIFICATE-----\n", "")

        f = open(os.path.join(app.config.get("OTS_CA_FOLDER"), "ca.pem"), "r")
        cert = f.read()
        f.close()

        cert = cert.replace("-----BEGIN CERTIFICATE-----\n", "")
        cert = cert.replace("\n-----END CERTIFICATE-----\n", "")

        # iTAK expects a JSON response but with the Content-Type header set to text/plain for some reason
        if (
            request.headers.get("Accept") == "text/plain"
            or request.headers.get("Accept") == "application/json"
            or request.headers.get("Accept") == "*/*"
            or not request.headers.get("Accept")
        ):
            response = {"signedCert": signed_csr, "ca0": cert, "ca1": cert}
        else:
            enrollment = Element("enrollment")
            signed_cert = SubElement(enrollment, "signedCert")
            signed_cert.text = signed_csr
            ca = SubElement(enrollment, "ca")
            ca.text = cert

            response = tostring(enrollment).decode("utf-8")
            response = '<?xml version="1.0" encoding="UTF-8"?>\n' + response

        username, password = (
            base64.b64decode(request.headers.get("Authorization").split(" ")[-1].encode("utf-8"))
            .decode("utf-8")
            .split(":")
        )
        username = bleach.clean(username)
        user = app.security.datastore.find_user(username=username)

        try:
            if uid:
                eud = EUD()
                eud.uid = uid
                eud.user_id = user.id

                db.session.add(eud)
                db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()
            eud = db.session.execute(db.session.query(EUD).filter_by(uid=uid)).first()[0]
            if user and not eud.user_id:
                eud.user_id = user.id
                db.session.add(eud)
                db.session.commit()

        try:
            certificate = Certificate()
            if uid:
                certificate.common_name = common_name
                certificate.eud_uid = uid
                certificate.callsign = eud.callsign
                certificate.expiration_date = datetime.datetime.today() + datetime.timedelta(
                    days=app.config.get("OTS_CA_EXPIRATION_TIME")
                )
                certificate.server_address = urlparse(request.url_root).hostname
                certificate.server_port = app.config.get("OTS_MARTI_HTTPS_PORT")
                certificate.truststore_filename = os.path.join(
                    app.config.get("OTS_CA_FOLDER"), "truststore-root.p12"
                )
                certificate.user_cert_filename = os.path.join(
                    app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pem"
                )
                certificate.csr = os.path.join(
                    app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr"
                )
                certificate.cert_password = app.config.get("OTS_CA_PASSWORD")

                db.session.add(certificate)
                db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()
            if uid:
                certificate = db.session.execute(
                    db.session.query(Certificate).filter_by(eud_uid=eud.uid)
                ).scalar_one()
                certificate.common_name = common_name
                certificate.callsign = eud.callsign
                certificate.expiration_date = datetime.datetime.today() + datetime.timedelta(
                    days=app.config.get("OTS_CA_EXPIRATION_TIME")
                )
                certificate.server_address = urlparse(request.url_root).hostname
                certificate.server_port = app.config.get("OTS_MARTI_HTTPS_PORT")
                certificate.truststore_filename = os.path.join(
                    app.config.get("OTS_CA_FOLDER"), "truststore-root.p12"
                )
                certificate.user_cert_filename = os.path.join(
                    app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pem"
                )
                certificate.csr = os.path.join(
                    app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr"
                )
                certificate.cert_password = app.config.get("OTS_CA_PASSWORD")

                db.session.commit()

        if request.headers.get("Accept") == "text/plain":
            return (
                response,
                200,
                {"Content-Type": "text/plain", "Content-Encoding": "charset=UTF-8"},
            )
        elif (
            request.headers.get("Accept") == "application/json"
            or request.headers.get("Accept") == "*/*"
        ):
            return jsonify(response)
        else:
            return (
                response,
                200,
                {"Content-Type": "application/xml", "Content-Encoding": "charset=UTF-8"},
            )
    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({"error": str(e)}), 500



from flask import Blueprint, jsonify, request

from opentakserver.extensions import db, logger
from opentakserver.models.EUD import EUD

contacts_api = Blueprint("contacts_api", __name__)


@contacts_api.route("/Marti/api/contacts/all")
def get_all_contacts():
    euds = db.session.execute(db.session.query(EUD)).all()

    response = []

    for eud in euds:
        eud = eud[0]
        team_name = eud.team.name if eud.team else "Cyan"
        team_role = eud.team_role or "Team Member"
        username = eud.user.username if eud.user else ""
        response.append(
            {
                "filterGroups": [],
                "notes": username,
                "callsign": eud.callsign,
                "team": team_name,
                "role": team_role,
                "takv": f"{eud.platform} {eud.version}",
                "uid": eud.uid,
            }
        )

    return jsonify(response)



"""Meshtastic Support

Revision ID: d91957bb59a0
Revises: 4c7909c34d4e
Create Date: 2024-05-24 04:16:20.765489

"""

import logging

import sqlalchemy as sa
from alembic import op

logger = logging.getLogger("OpenTAKServer")
logger.disabled = False


# revision identifiers, used by Alembic.
revision = "d91957bb59a0"
down_revision = "4c7909c34d4e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        op.create_table(
            "iconsets",
            sa.Column("id", sa.INTEGER(), nullable=False),
            sa.Column("name", sa.String(length=255), nullable=True),
            sa.Column("uid", sa.String(length=255), nullable=True),
            sa.Column("version", sa.Integer(), nullable=True),
            sa.Column("defaultFriendly", sa.String(length=255), nullable=True),
            sa.Column("defaultHostile", sa.String(length=255), nullable=True),
            sa.Column("defaultNeutral", sa.String(length=255), nullable=True),
            sa.Column("defaultUnknown", sa.String(length=255), nullable=True),
            sa.Column("selectedGroup", sa.String(length=255), nullable=True),
            sa.PrimaryKeyConstraint("id"),
        )
    except sa.exc.OperationalError:
        logger.warning("iconsets table already exists, skipping")

    try:
        op.create_table(
            "meshtastic_channels",
            sa.Column("id", sa.Integer(), nullable=False),
            sa.Column("psk", sa.String(length=255), nullable=True),
            sa.Column("name", sa.String(length=255), nullable=True),
            sa.Column("uplink_enabled", sa.Boolean(), nullable=False),
            sa.Column("downlink_enabled", sa.Boolean(), nullable=False),
            sa.Column("position_precision", sa.Integer(), nullable=False),
            sa.Column("lora_region", sa.Integer(), nullable=False),
            sa.Column("lora_hop_limit", sa.Integer(), nullable=False),
            sa.Column("lora_tx_enabled", sa.Boolean(), nullable=False),
            sa.Column("lora_tx_power", sa.Integer(), nullable=False),
            sa.Column("lora_sx126x_rx_boosted_gain", sa.Boolean(), nullable=False),
            sa.Column("modem_preset", sa.Integer(), nullable=False),
            sa.Column("url", sa.String(length=255), nullable=False),
            sa.PrimaryKeyConstraint("id"),
            sa.UniqueConstraint("url"),
        )

        with op.batch_alter_table("euds", schema=None) as batch_op:
            batch_op.add_column(sa.Column("meshtastic_id", sa.Integer(), nullable=True))
            batch_op.add_column(
                sa.Column("meshtastic_macaddr", sa.String(length=255), nullable=True)
            )
    except sa.exc.OperationalError:
        logger.warning("Meshtastic table already exists, skipping")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("euds", schema=None) as batch_op:
        batch_op.drop_column("meshtastic_macaddr")
        batch_op.drop_column("meshtastic_id")

    op.drop_table("meshtastic_channels")
    op.drop_table("iconsets")
    # ### end Alembic commands ###



from dataclasses import dataclass

from sqlalchemy import ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db


@dataclass
class Team(db.Model):
    __tablename__ = "teams"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String(255), unique=True)
    chatroom_id: Mapped[str] = mapped_column(String(255), ForeignKey("chatrooms.id"), nullable=True)
    euds = relationship("EUD", back_populates="team")
    chatroom = relationship("Chatroom", back_populates="team")
    mission_invitations = relationship("MissionInvitation", back_populates="team")

    colors = {
        "Cyan": "#00FFFF",
        "White": "#000000",
        "Yellow": "#FFFF00",
        "Orange": "#FFA500",
        "Magenta": "#FF00FF",
        "Red": "#FF0000",
        "Maroon": "#800000",
        "Purple": "#800080",
        "Dark Blue": "#00008B",
        "Blue": "#0000FF",
        "Teal": "#008080",
        "Green": "#00FF00",
        "Dark Green": "#228B22",
        "Brown": "#964B00",
        "Gray": "#808080",
    }

    def get_team_color(self):
        return self.colors[self.name]

    def serialize(self):
        return {
            "name": self.name,
        }

    def to_json(self):
        return {
            "name": self.name,
            "chatroom": self.chatroom.to_json() if self.chatroom else None,
            "euds": [eud.to_json() for eud in self.euds] if self.euds else None,
            "color": self.colors[self.name],
        }



"""Fixes for the eud_stats table

Revision ID: c9fa07339363
Revises: 85dbe0a7dcc1
Create Date: 2025-03-21 15:01:54.874938

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c9fa07339363"
down_revision = "85dbe0a7dcc1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("eud_stats", schema=None) as batch_op:
        batch_op.alter_column(
            "heap_free_size",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "storage_total",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "deviceDataRx",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "heap_max_size",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "storage_available",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "deviceDataTx",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("eud_stats", schema=None) as batch_op:
        batch_op.alter_column(
            "deviceDataTx",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "storage_available",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "heap_max_size",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "deviceDataRx",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "storage_total",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "heap_free_size",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



"""Added mission_uids table

Revision ID: 807c3ca8e7d0
Revises: bde915ea136e
Create Date: 2024-10-11 16:53:44.954118

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision = "807c3ca8e7d0"
down_revision = "bde915ea136e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "mission_uids",
        sa.Column("uid", sa.String(length=255), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=True),
        sa.Column("timestamp", sa.DateTime(), nullable=True),
        sa.Column("creator_uid", sa.String(length=255), nullable=True),
        sa.Column("cot_type", sa.String(length=255), nullable=True),
        sa.Column("callsign", sa.String(length=255), nullable=True),
        sa.Column("iconset_path", sa.String(length=255), nullable=True),
        sa.Column("color", sa.String(length=255), nullable=True),
        sa.Column("latitude", sa.Float(), nullable=True),
        sa.Column("longitude", sa.Float(), nullable=True),
        sa.ForeignKeyConstraint(
            ["mission_name"],
            ["missions.name"],
        ),
        sa.PrimaryKeyConstraint("uid"),
    )
    op.drop_table("mission_logs")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "mission_logs",
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.Column("content", sa.String(length=255), nullable=False),
        sa.Column("creator_uid", sa.String(length=255), nullable=False),
        sa.Column("entry_uid", sa.String(length=255), nullable=False),
        sa.Column("server_time", sa.DATETIME(), nullable=False),
        sa.Column("dtg", sa.DATETIME(), nullable=False),
        sa.Column("created", sa.DATETIME(), nullable=False),
        sa.Column("keywords", sqlite.JSON(), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=True),
        sa.Column("content_hash", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.drop_table("mission_uids")
    # ### end Alembic commands ###



import datetime
import uuid
from dataclasses import dataclass
from xml.etree.ElementTree import Element, SubElement

from sqlalchemy import JSON, Boolean, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.MissionChange import MissionChange


@dataclass
class MissionLogEntry(db.Model):
    __tablename__ = "mission_logs"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    content: Mapped[str] = mapped_column(String(255))
    creator_uid: Mapped[str] = mapped_column(String(255))
    entry_uid: Mapped[str] = mapped_column(String(255), default=str(uuid.uuid4()))
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"))
    server_time: Mapped[datetime] = mapped_column(
        DateTime, default=datetime.datetime.now(datetime.timezone.utc)
    )
    dtg: Mapped[datetime] = mapped_column(
        DateTime, default=datetime.datetime.now(datetime.timezone.utc)
    )
    created: Mapped[datetime] = mapped_column(
        DateTime, default=datetime.datetime.now(datetime.timezone.utc)
    )
    # Leaving content_hash in order to not break existing DBs where this field is populated
    content_hash: Mapped[str] = mapped_column(String(255), nullable=True)
    content_hashes: Mapped[JSON] = mapped_column(JSON, default=[], nullable=True)
    keywords: Mapped[JSON] = mapped_column(JSON, default=[])
    mission = relationship("Mission", back_populates="mission_logs")

    def serialize(self):
        return {
            "content": self.content,
            "creator_uid": self.creator_uid,
            "entry_uid": self.entry_uid,
            "mission_names": self.mission_name,
            "server_time": self.server_time,
            "dtg": self.dtg,
            "created": self.created,
            "content_hashes": self.content_hashes,
            "keywords": self.keywords,
        }

    def to_json(self):
        return {
            "id": self.entry_uid,
            "content": self.content,
            "creatorUid": self.creator_uid,
            "entryUid": self.entry_uid,
            "missionNames": [self.mission_name],
            "servertime": iso8601_string_from_datetime(self.server_time),
            "dtg": iso8601_string_from_datetime(self.dtg),
            "created": iso8601_string_from_datetime(self.created),
            "contentHashes": self.content_hashes if self.content_hash else [],
            "keywords": self.keywords if self.keywords is not None else [],
        }

    def generate_cot(self) -> Element:
        event = Element(
            "event",
            {
                "how": "h-g-i-g-o",
                "type": "t-x-m-c-l",
                "version": "2.0",
                "uid": str(uuid.uuid4()),
                "start": iso8601_string_from_datetime(self.dtg),
                "time": iso8601_string_from_datetime(self.server_time),
                "stale": iso8601_string_from_datetime(self.dtg + datetime.timedelta(minutes=2)),
                "access": "Undefined",
            },
        )

        SubElement(
            event,
            "point",
            {"ce": "9999999", "le": "9999999", "hae": "0.0", "lat": "0.0", "lon": "0.0"},
        )

        detail = SubElement(event, "detail")

        SubElement(
            detail,
            "mission",
            {
                "type": MissionChange.CHANGE,
                "tool": "public",
                "name": self.mission.name,
                "guid": self.mission.guid,
                "authorUid": self.creator_uid,
            },
        )

        return event



import os
import uuid
from dataclasses import dataclass
from datetime import datetime

from sqlalchemy import Boolean, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class DataPackage(db.Model):
    __tablename__ = "data_packages"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    filename: Mapped[str] = mapped_column(String(255), unique=True)
    hash: Mapped[str] = mapped_column(String(255), unique=True)
    creator_uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True
    )
    submission_time: Mapped[datetime] = mapped_column(DateTime)
    submission_user: Mapped[int] = mapped_column(Integer, ForeignKey("user.id"), nullable=True)
    keywords: Mapped[str] = mapped_column(String(255), nullable=True)
    mime_type: Mapped[str] = mapped_column(String(255))
    size: Mapped[int] = mapped_column(Integer)
    tool: Mapped[str] = mapped_column(String(255), nullable=True)
    expiration: Mapped[str] = mapped_column(String(255), nullable=True)
    install_on_enrollment: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    install_on_connection: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    eud: Mapped["EUD"] = relationship(back_populates="data_packages")
    certificate = relationship("Certificate", back_populates="data_package", uselist=False)
    user = relationship("User", back_populates="data_packages")

    def serialize(self):
        return {
            "filename": self.filename,
            "hash": self.hash,
            "creator_uid": self.creator_uid,
            "submission_time": self.submission_time,
            "submission_user": self.submission_user,
            "keywords": self.keywords,
            "mime_type": self.mime_type,
            "size": self.size,
            "tool": self.tool,
            "expiration": self.expiration,
            "install_on_enrollment": self.install_on_enrollment,
            "install_on_connection": self.install_on_connection,
        }

    def to_json(self, include_eud=True):
        return {
            "filename": self.filename,
            "hash": self.hash,
            "creator_uid": self.creator_uid,
            "submission_time": iso8601_string_from_datetime(self.submission_time),
            "submission_user": self.user.username if self.user else None,
            "keywords": self.keywords,
            "mime_type": self.mime_type,
            "size": self.size,
            "tool": self.tool,
            "expiration": self.expiration,
            "eud": self.eud.to_json(False) if include_eud and self.eud else None,
            "install_on_enrollment": self.install_on_enrollment,
            "install_on_connection": self.install_on_connection,
        }



"""Added distro column to plugin table

Revision ID: b0c97a4ce0df
Revises: f310052349cb
Create Date: 2025-05-02 21:27:56.708927

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b0c97a4ce0df"
down_revision = "f310052349cb"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("plugins", schema=None) as batch_op:
        batch_op.add_column(sa.Column("distro", sa.String(length=255), nullable=False))
        batch_op.create_unique_constraint("distro", ["distro"])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("plugins", schema=None) as batch_op:
        batch_op.drop_constraint("distro", type_="unique")
        batch_op.drop_column("distro")

    # ### end Alembic commands ###
